from __future__ import annotations

from typing import TYPE_CHECKING, Awaitable, Callable

from infrahub.components import ComponentType
from infrahub.exceptions import InitializationError
from infrahub.log import get_logger
from infrahub.message_bus.messages import ROUTING_KEY_MAP

from .adapters.event import InfrahubEventService
from .adapters.http.httpx import HttpxAdapter
from .adapters.workflow.worker import WorkflowWorkerExecution
from .scheduler import InfrahubScheduler

if TYPE_CHECKING:
    from infrahub_sdk import InfrahubClient

    from infrahub.database import InfrahubDatabase
    from infrahub.message_bus import InfrahubMessage
    from infrahub.message_bus.types import MessageTTL

    from .adapters.cache import InfrahubCache
    from .adapters.http import InfrahubHTTP
    from .adapters.message_bus import InfrahubMessageBus
    from .adapters.workflow import InfrahubWorkflow
    from .component import InfrahubComponent
    from .protocols import InfrahubLogger


class InfrahubServices:
    _cache: InfrahubCache | None
    _client: InfrahubClient | None
    _database: InfrahubDatabase | None
    _message_bus: InfrahubMessageBus | None
    _workflow: InfrahubWorkflow | None
    _component: InfrahubComponent | None

    log: InfrahubLogger
    component_type: ComponentType
    http: InfrahubHTTP
    event: InfrahubEventService
    scheduler: InfrahubScheduler

    def __init__(
        self,
        log: InfrahubLogger,
        component_type: ComponentType,
        http: InfrahubHTTP,
        event: InfrahubEventService,
        scheduler: InfrahubScheduler,
        cache: InfrahubCache | None = None,
        client: InfrahubClient | None = None,
        database: InfrahubDatabase | None = None,
        message_bus: InfrahubMessageBus | None = None,
        workflow: InfrahubWorkflow | None = None,
        component: InfrahubComponent | None = None,
    ) -> None:
        """
        This method should not be called directly, use `new` instead for a proper initialization.
        """

        self._cache = cache
        self._client = client
        self._database = database
        self._message_bus = message_bus
        self._workflow = workflow
        self._component = component
        self.log = log
        self.component_type = component_type
        self.http = http
        self.event = event
        self.scheduler = scheduler

    @classmethod
    async def new(
        cls,
        cache: InfrahubCache | None = None,
        client: InfrahubClient | None = None,
        database: InfrahubDatabase | None = None,
        event: InfrahubEventService | None = None,
        message_bus: InfrahubMessageBus | None = None,
        workflow: InfrahubWorkflow | None = None,
        log: InfrahubLogger | None = None,
        component: InfrahubComponent | None = None,
        component_type: ComponentType | None = None,
        http: InfrahubHTTP | None = None,
    ) -> InfrahubServices:
        """
        Instantiate InfrahubServices object, and finalize initializations of underlying services having a circular
        dependency with InfrahubServices.
        """

        component_type = component_type or ComponentType.NONE

        scheduler = InfrahubScheduler(component_type)
        service = cls(
            cache=cache,
            client=client,
            database=database,
            message_bus=message_bus,
            workflow=workflow,
            log=log or get_logger(),
            component=component,
            component_type=component_type,
            scheduler=scheduler,
            event=event or InfrahubEventService(message_bus),
            http=http or HttpxAdapter(),
        )

        # This circular dependency could be removed if InfrahubScheduler only depends on what it needs.
        scheduler.service = service

        return service

    async def initialize_workflow(self, is_initial_setup: bool = False) -> None:
        if self.workflow is not None and isinstance(self.workflow, WorkflowWorkerExecution):
            assert self.component is not None
            # Ideally `WorkflowWorkerExecution.initialize` would be directly part of WorkflowWorkerExecution
            # constructor but this requires some redesign as it depends on InfrahubComponent which is instantiated
            # after workflow instantiation.
            await self.component.refresh_heartbeat()
            is_primary = await self.component.is_primary_gunicorn_worker()
            await self.workflow.initialize(component_is_primary_server=is_primary, is_initial_setup=is_initial_setup)

    @property
    def component(self) -> InfrahubComponent:
        if not self._component:
            raise InitializationError("Service is not initialized with a component")

        return self._component

    @property
    def message_bus(self) -> InfrahubMessageBus:
        if not self._message_bus:
            raise InitializationError("Service is not initialized with a message bus")

        return self._message_bus

    @property
    def workflow(self) -> InfrahubWorkflow:
        if not self._workflow:
            raise InitializationError("Service is not initialized with a workflow")

        return self._workflow

    @property
    def cache(self) -> InfrahubCache:
        if not self._cache:
            raise InitializationError("Service is not initialized with a cache")

        return self._cache

    @property
    def client(self) -> InfrahubClient:
        if not self._client:
            raise InitializationError("Service is not initialized with a client")

        return self._client

    @property
    def database(self) -> InfrahubDatabase:
        if not self._database:
            raise InitializationError("Service is not initialized with a database")

        return self._database

    async def shutdown(self) -> None:
        await self.scheduler.shutdown()
        await self.message_bus.shutdown()
        if self._cache is not None:
            await self._cache.close_connection()

    async def send(self, message: InfrahubMessage, delay: MessageTTL | None = None, is_retry: bool = False) -> None:
        routing_key = ROUTING_KEY_MAP.get(type(message))
        if not routing_key:
            raise ValueError("Unable to determine routing key")
        await self.message_bus.publish(message, routing_key=routing_key, delay=delay, is_retry=is_retry)


ServiceFunction = Callable[[InfrahubServices], Awaitable[None]]



from typing import Any, TypeVar, cast

from infrahub_sdk.client import InfrahubClient
from infrahub_sdk.protocols import CoreValidator

from infrahub.context import InfrahubContext
from infrahub.core.constants import ValidatorConclusion, ValidatorState
from infrahub.workers.dependencies import get_event_service

from .events import send_start_validator

ValidatorType = TypeVar("ValidatorType", bound=CoreValidator)


async def start_validator[ValidatorType: CoreValidator](
    client: InfrahubClient,
    validator: CoreValidator | None,
    validator_type: type[ValidatorType],
    proposed_change: str,
    context: InfrahubContext,
    data: dict[str, Any],
) -> ValidatorType:
    if validator:
        validator.conclusion.value = ValidatorConclusion.UNKNOWN.value
        validator.state.value = ValidatorState.QUEUED.value
        validator.started_at.value = ""
        validator.completed_at.value = ""
        await validator.save()
        validator = cast("ValidatorType", validator)
    else:
        data["proposed_change"] = proposed_change
        client.request_context = context.to_request_context()
        validator = await client.create(kind=validator_type, data=data)
        await validator.save()

    event_service = await get_event_service()
    await send_start_validator(
        event_service=event_service, validator=validator, proposed_change_id=proposed_change, context=context
    )

    return validator



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field

from infrahub import config
from infrahub.exceptions import RPCError
from infrahub.log import set_log_data


class Meta(BaseModel):
    request_id: str = ""
    correlation_id: str | None = Field(default=None)
    reply_to: str | None = Field(default=None)
    initiator_id: str | None = Field(
        default=None, description="The worker identity of the initial sender of this message"
    )
    retry_count: int | None = Field(default=None, description="Indicates how many times this message has been retried.")
    headers: dict[str, Any] | None = Field(default=None)
    validator_execution_id: str | None = Field(
        default=None, description="Validator execution ID related to this message"
    )
    check_execution_id: str | None = Field(default=None, description="Check execution ID related to this message")
    priority: int = Field(default=3, description="Message Priority")
    expiration: int | None = Field(default=None, description="TTL before this message expires in seconds")

    @classmethod
    def default(cls) -> Meta:
        return cls()


class InfrahubMessage(BaseModel):
    """Base Model for messages"""

    meta: Meta = Field(default_factory=Meta.default, description="Meta properties for the message")

    def assign_meta(self, parent: InfrahubMessage) -> None:
        """Assign relevant meta properties from a parent message."""
        self.meta.request_id = parent.meta.request_id
        self.meta.initiator_id = parent.meta.initiator_id

    def assign_header(self, key: str, value: Any) -> None:
        self.meta.headers = self.meta.headers or {}
        self.meta.headers[key] = value

    def assign_priority(self, priority: int) -> None:
        self.meta.priority = priority

    def assign_expiration(self, expiration: int) -> None:
        self.meta.expiration = expiration

    def set_log_data(self, routing_key: str) -> None:
        set_log_data(key="routing_key", value=routing_key)
        if self.meta.request_id:
            set_log_data(key="request_id", value=self.meta.request_id)

    @property
    def reply_requested(self) -> bool:
        if self.meta.reply_to:
            return True
        return False

    @property
    def body(self) -> bytes:
        return self.model_dump_json(
            exclude={"meta": {"headers", "priority", "expiration"}, "value": True}, exclude_none=True
        ).encode("UTF-8")

    def increase_retry_count(self, count: int = 1) -> None:
        current_retry = self.meta.retry_count or 0
        self.meta.retry_count = current_retry + count

    @property
    def reached_max_retries(self) -> bool:
        if self.meta and self.meta.retry_count:
            return self.meta.retry_count >= config.SETTINGS.broker.maximum_message_retries
        return False


class InfrahubResponseData(BaseModel):
    pass


class InfrahubResponse(InfrahubMessage):
    """A response to an RPC request"""

    passed: bool = True
    routing_key: str
    data: dict | InfrahubResponseData = Field(default_factory=dict)  # type: ignore[arg-type]
    errors: list[str] = Field(default_factory=list)
    initial_message: dict | None = Field(
        default=None,
        description="Initial message in dict format, the primary goal of this field is to provide additional context when there is an error",
    )

    def raise_for_status(self) -> None:
        if self.passed:
            return

        raise RPCError(message=", ".join(self.errors or ["Unknown Error"]))


class RPCErrorResponse(InfrahubResponse):
    passed: bool = False
    routing_key: str = "rpc_error"



from infrahub.core.relationship.constraints.profiles_removal import RelationshipProfileRemovalConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class RelationshipProfileRemovalConstraintDependency(DependencyBuilder[RelationshipProfileRemovalConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipProfileRemovalConstraint:
        return RelationshipProfileRemovalConstraint(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.constants import PROFILES_RELATIONSHIP_NAME, MetadataOptions
from infrahub.core.manager import NodeManager
from infrahub.core.schema import NodeSchema, ProfileSchema
from infrahub.exceptions import ValidationError

from .interface import RelationshipManagerConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.relationship.model import Relationship, RelationshipManager
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase


class RelationshipProfileRemovalConstraint(RelationshipManagerConstraintInterface):
    """Constraint that validates removing profiles from a node doesn't violate required relationships.

    This runs in two cases:
    1. When a node's `profiles` relationship is changed.
    2. When a profile's `related_nodes` relationship is changed.

    In both cases, it will perform checks only if peers are being removed from the relationship being changed.
    """

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch
        self.schema_branch = registry.schema.get_schema_branch(branch.name if branch else registry.default_branch)

    def _get_required_attributes_names(self, schema: NodeSchema) -> set[str]:
        attr_names: set[str] = set()
        for attr_schema in schema.attributes:
            if attr_schema.support_profiles and not attr_schema.optional:
                attr_names.add(attr_schema.name)
        return attr_names

    def _get_required_relationship_names(self, schema: NodeSchema) -> set[str]:
        rel_names: set[str] = set()
        for rel_schema in schema.relationships:
            if rel_schema.support_profiles and not rel_schema.optional:
                rel_names.add(rel_schema.name)
        return rel_names

    async def _validate_profile_removal(
        self, node: Node, profile_id: str, required_attr_names: set[str], required_rel_names: set[str]
    ) -> None:
        for attr_name in required_attr_names:
            attr = node.get_attribute(name=attr_name)
            if attr.is_from_profile:
                source = await attr.get_source(db=self.db)
                if source and source.id == profile_id:
                    node_display_label = await node.get_display_label(db=self.db)
                    node_reference = f"node '{node_display_label}' (ID: {node.get_id()})"
                    raise ValidationError(
                        f"Cannot remove profile '{profile_id}' because {node_reference} "
                        f"inherits required attribute '{attr_name}' from this profile."
                    )

        for rel_name in required_rel_names:
            rel_manager = node.get_relationship(name=rel_name)

            relationships: list[Relationship] = await rel_manager.get_relationships(db=self.db)
            for rel in relationships:
                if rel.is_from_profile:
                    source = await rel.get_source(db=self.db)
                    if source and source.id == profile_id:
                        node_display_label = await node.get_display_label(db=self.db)
                        node_reference = f"node '{node_display_label}' (ID: {node.get_id()})"
                        raise ValidationError(
                            f"Cannot remove profile '{profile_id}' because {node_reference} "
                            f"inherits required relationship '{rel_name}' from this profile."
                        )

    async def _check_node_profiles_removal(
        self, relm: RelationshipManager, node_schema: NodeSchema, node: Node
    ) -> None:
        required_attr_names = self._get_required_attributes_names(schema=node_schema)
        required_rel_names = self._get_required_relationship_names(schema=node_schema)
        if not required_attr_names and not required_rel_names:
            return

        relm_update_details = await relm.fetch_relationship_ids(db=self.db, force_refresh=False)
        if not relm_update_details.peer_ids_present_database_only:
            return

        if required_attr_names:
            # Required to get source for attributes
            node = await NodeManager.get_one(
                db=self.db, branch=self.branch, id=node.get_id(), include_metadata=MetadataOptions.SOURCE
            )

        for profile_id in relm_update_details.peer_ids_present_database_only:
            await self._validate_profile_removal(
                node=node,
                profile_id=profile_id,
                required_attr_names=required_attr_names,
                required_rel_names=required_rel_names,
            )

    async def _check_profile_related_nodes_removal(
        self, relm: RelationshipManager, profile_schema: ProfileSchema, profile: Node
    ) -> None:
        relm_update_details = await relm.fetch_relationship_ids(db=self.db, force_refresh=False)
        if not relm_update_details.peer_ids_present_database_only:
            return

        target_kind = profile_schema.get_relationship(name="related_nodes").peer
        target_schema = self.schema_branch.get_node(name=target_kind, duplicate=False)

        required_attr_names = self._get_required_attributes_names(schema=target_schema)
        required_rel_names = self._get_required_relationship_names(schema=target_schema)
        if not required_attr_names and not required_rel_names:
            return

        nodes = await NodeManager.get_many(
            db=self.db,
            branch=self.branch,
            ids=relm_update_details.peer_ids_present_database_only,
            include_metadata=MetadataOptions.SOURCE,
        )
        for node in nodes.values():
            await self._validate_profile_removal(
                node=node,
                profile_id=profile.get_id(),
                required_attr_names=required_attr_names,
                required_rel_names=required_rel_names,
            )

    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes, node: Node) -> None:
        if relm.name == PROFILES_RELATIONSHIP_NAME and isinstance(node_schema, NodeSchema):
            await self._check_node_profiles_removal(relm=relm, node_schema=node_schema, node=node)
            return

        if relm.name == "related_nodes" and isinstance(node_schema, ProfileSchema):
            await self._check_profile_related_nodes_removal(relm=relm, profile_schema=node_schema, profile=node)
            return

    async def validate_profile_deletion(self, profile: Node, profile_schema: ProfileSchema) -> None:
        related_nodes_rels = await profile.related_nodes.get_relationships(db=self.db)  # type: ignore[attr-defined]
        related_node_ids = [rel.peer_id for rel in related_nodes_rels if rel.peer_id]

        if not related_node_ids:
            return

        target_kind = profile_schema.get_relationship(name="related_nodes").peer
        target_schema = self.schema_branch.get_node(name=target_kind, duplicate=False)

        required_attr_names = self._get_required_attributes_names(schema=target_schema)
        required_rel_names = self._get_required_relationship_names(schema=target_schema)
        if not required_attr_names and not required_rel_names:
            return

        nodes = await NodeManager.get_many(
            db=self.db, branch=self.branch, ids=related_node_ids, include_metadata=MetadataOptions.SOURCE
        )
        for node in nodes.values():
            await self._validate_profile_removal(
                node=node,
                profile_id=profile.get_id(),
                required_attr_names=required_attr_names,
                required_rel_names=required_rel_names,
            )



from infrahub.events.branch_action import BranchDeletedEvent
from infrahub.events.schema_action import SchemaUpdatedEvent
from infrahub.trigger.models import BuiltinTriggerDefinition, EventTrigger, ExecuteWorkflow
from infrahub.workflows.catalogue import PROFILE_REFRESH_SETUP

TRIGGER_PROFILE_REFRESH_SETUP = BuiltinTriggerDefinition(
    name="profile-refresh-setup-all",
    trigger=EventTrigger(events={SchemaUpdatedEvent.event_name, BranchDeletedEvent.event_name}),
    actions=[
        ExecuteWorkflow(
            workflow=PROFILE_REFRESH_SETUP,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "event_name": "{{ event.event }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['context'] | tojson }}"},
                },
            },
        ),
    ],
)



from __future__ import annotations

from typing import TYPE_CHECKING, NoReturn

from fastapi import APIRouter, Depends
from fastapi.openapi.docs import (
    get_redoc_html,
    get_swagger_ui_html,
)
from starlette.responses import HTMLResponse  # noqa: TC002

from infrahub.api import (
    artifact,
    auth,
    diff,
    file,
    internal,
    menu,
    oauth2,
    oidc,
    query,
    schema,
    transformation,
)
from infrahub.api.dependencies import get_current_user
from infrahub.api.storage import storage
from infrahub.exceptions import ResourceNotFoundError

if TYPE_CHECKING:
    from infrahub.auth import AccountSession

router = APIRouter(prefix="/api")

router.include_router(artifact.router)
router.include_router(auth.router)
router.include_router(diff.router)
router.include_router(file.router)
router.include_router(internal.router)
router.include_router(menu.router)
router.include_router(oauth2.router)
router.include_router(oidc.router)
router.include_router(query.router)
router.include_router(schema.router)
router.include_router(storage.router)
router.include_router(transformation.router)


@router.get("/docs", include_in_schema=False)
async def custom_swagger_ui_html(
    _: AccountSession = Depends(get_current_user),
) -> HTMLResponse:
    return get_swagger_ui_html(
        openapi_url="/api/openapi.json",
        title="Infrahub - Swagger UI",
        swagger_js_url="/api-static/swagger-ui-bundle.js",
        swagger_css_url="/api-static/swagger-ui.css",
    )


@router.get("/redoc", include_in_schema=False)
async def redoc_html(_: AccountSession = Depends(get_current_user)) -> HTMLResponse:
    return get_redoc_html(
        openapi_url="/api/openapi.json",
        title="Infrahub - ReDoc",
        redoc_js_url="/api-static/redoc.standalone.js",
    )


@router.api_route(
    "/{rest_of_path:path}",
    methods=["GET", "POST", "PUT", "DELETE", "OPTIONS"],
    include_in_schema=False,
    response_model=None,
)
async def not_found(rest_of_path: str) -> NoReturn:
    """Used to avoid having the mounting of the React App mask 404 errors."""
    raise ResourceNotFoundError(
        message=f"The requested endpoint /api/{rest_of_path} does not exist",
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub import config
from infrahub.core.constants import PathType, RelationshipKind
from infrahub.core.path import DataPath, GroupedDataPaths
from infrahub.core.schema import GenericSchema

from ..interface import ConstraintCheckerInterface
from ..shared import RelationshipSchemaValidatorQuery

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema.relationship_schema import RelationshipSchema
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class RelationshipPeerUpdateValidatorQuery(RelationshipSchemaValidatorQuery):
    name = "relationship_constraints_peer_validator"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        peer_schema = db.schema.get(name=self.relationship_schema.peer, branch=self.branch, duplicate=False)
        allowed_peer_kinds = [peer_schema.kind]
        if isinstance(peer_schema, GenericSchema):
            allowed_peer_kinds += peer_schema.used_by

        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)

        self.params["relationship_id"] = self.relationship_schema.identifier
        self.params["allowed_peer_kinds"] = allowed_peer_kinds

        # ruff: noqa: E501
        query = """
        MATCH (n:%(node_kind)s)
        CALL (n) {
            MATCH path = (root:Root)<-[rroot:IS_PART_OF]-(n)
            WHERE all(r in relationships(path) WHERE %(branch_filter)s)
            RETURN path as full_path, n as active_node
            ORDER BY rroot.branch_level DESC, rroot.from DESC
            LIMIT 1
        }
        WITH full_path, active_node
        WHERE all(r in relationships(full_path) WHERE r.status = "active")
        CALL (active_node) {
            MATCH path = (active_node)-[rrel1:IS_RELATED]-(rel:Relationship { name: $relationship_id })-[rrel2:IS_RELATED]-(peer:Node)
            WHERE all(
                r in relationships(path)
                WHERE (%(branch_filter)s)
            )
            RETURN
                path as relationship_path,
                active_node as start_node,
                peer as peer_node,
                rrel1.branch_level + rrel2.branch_level AS branch_level_sum,
                [rrel1.from, rrel2.from] as from_times,
                // used as tiebreaker for updated relationships that were deleted and added at the same microsecond
                reduce(active_count = 0, r in relationships(path) | active_count + (CASE r.status WHEN "active" THEN 1 ELSE 0 END)) AS active_relationship_count,
                (CASE WHEN rrel1.branch_level > rrel2.branch_level THEN rrel1.branch ELSE rrel2.branch END) as deepest_branch_name
        }
        WITH
            collect([branch_level_sum, from_times, active_relationship_count, relationship_path, deepest_branch_name]) as enriched_paths,
            start_node,
            peer_node
        CALL (enriched_paths, peer_node) {
            UNWIND enriched_paths as path_to_check
            RETURN path_to_check[3] as current_path, path_to_check[4] as branch_name, peer_node as current_peer
            ORDER BY
                path_to_check[0] DESC,
                path_to_check[1][1] DESC,
                path_to_check[1][0] DESC,
                path_to_check[2] DESC
            LIMIT 1
        }
        WITH start_node, current_peer, branch_name, current_path
        WHERE all(r in relationships(current_path) WHERE r.status = "active")
        AND NOT any(label IN LABELS(current_peer) WHERE label IN $allowed_peer_kinds)
        """ % {"branch_filter": branch_filter, "node_kind": self.node_schema.kind}

        self.add_to_query(query)
        self.return_labels = ["start_node.uuid", "branch_name", "current_peer.uuid"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("branch_name")),
                    path_type=PathType.NODE,
                    node_id=str(result.get("start_node.uuid")),
                    field_name=self.relationship_schema.name,
                    peer_id=str(result.get("current_peer.uuid")),
                    kind=self.node_schema.kind,
                )
            )

        return grouped_data_paths


class RelationshipPeerChecker(ConstraintCheckerInterface):
    query_classes = [RelationshipPeerUpdateValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "relationship.peer.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []

        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db, branch=self.branch, node_schema=request.node_schema, schema_path=request.schema_path
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list


class RelationshipPeerParentValidatorQuery(RelationshipSchemaValidatorQuery):
    name = "relationship_constraints_peer_parent_validator"

    def __init__(
        self,
        relationship: RelationshipSchema,
        parent_relationship: RelationshipSchema,
        peer_parent_relationship: RelationshipSchema,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        self.relationship = relationship
        self.parent_relationship = parent_relationship
        self.peer_parent_relationship = peer_parent_relationship

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)
        self.params["peer_relationship_id"] = self.relationship.identifier
        self.params["parent_relationship_id"] = self.parent_relationship.identifier
        self.params["peer_parent_relationship_id"] = self.peer_parent_relationship.identifier

        parent_arrows = self.parent_relationship.get_query_arrows()
        parent_match = (
            "MATCH (active_node)%(lstart)s[r1:IS_RELATED]%(lend)s"
            "(rel:Relationship { name: $parent_relationship_id })%(rstart)s[r2:IS_RELATED]%(rend)s(parent:Node)"
        ) % {
            "lstart": parent_arrows.left.start,
            "lend": parent_arrows.left.end,
            "rstart": parent_arrows.right.start,
            "rend": parent_arrows.right.end,
        }

        peer_parent_arrows = self.relationship.get_query_arrows()
        peer_match = (
            "MATCH (active_node)%(lstart)s[r1:IS_RELATED]%(lend)s"
            "(r:Relationship {name: $peer_relationship_id })%(rstart)s[r2:IS_RELATED]%(rend)s(peer:Node)"
        ) % {
            "lstart": peer_parent_arrows.left.start,
            "lend": peer_parent_arrows.left.end,
            "rstart": peer_parent_arrows.right.start,
            "rend": peer_parent_arrows.right.end,
        }

        peer_parent_arrows = self.peer_parent_relationship.get_query_arrows()
        peer_parent_match = (
            "MATCH (peer:Node)%(lstart)s[r1:IS_RELATED]%(lend)s"
            "(r:Relationship {name: $peer_parent_relationship_id})%(rstart)s[r2:IS_RELATED]%(rend)s(peer_parent:Node)"
        ) % {
            "lstart": peer_parent_arrows.left.start,
            "lend": peer_parent_arrows.left.end,
            "rstart": peer_parent_arrows.right.start,
            "rend": peer_parent_arrows.right.end,
        }

        query = """
        MATCH (n:%(node_kind)s)
        CALL (n) {
            MATCH path = (root:Root)<-[r:IS_PART_OF]-(n)
            WHERE %(branch_filter)s
            RETURN n as active_node, r.status = "active" AS is_active
            ORDER BY r.branch_level DESC, r.from DESC
            LIMIT 1
        }
        WITH active_node, is_active
        WHERE is_active = TRUE
        %(parent_match)s
        WHERE all(r in [r1, r2] WHERE %(branch_filter)s AND r.status = "active")
        CALL (active_node) {
            %(peer_match)s
            WITH DISTINCT active_node, peer
            %(peer_match)s
            WHERE all(r in [r1, r2] WHERE %(branch_filter)s)
            WITH peer, r1.status = "active" AND r2.status = "active" AS is_active
            ORDER BY peer.uuid, r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC, is_active DESC
            WITH peer, head(collect(is_active)) AS is_active
            WHERE is_active = TRUE
            RETURN peer
        }
        CALL (peer) {
            %(peer_parent_match)s
            WHERE all(r IN [r1, r2] WHERE %(branch_filter)s)
            WITH peer_parent, r1, r2, r1.status = "active" AND r2.status = "active" AS is_active
            WITH peer_parent, r1.branch AS branch_name, is_active
            ORDER BY r1.branch_level DESC, r2.branch_level DESC, r1.from DESC, r2.from DESC, is_active DESC
            LIMIT 1
            WITH peer_parent, branch_name
            WHERE is_active = TRUE
            RETURN peer_parent, branch_name
        }
        WITH DISTINCT active_node, parent, peer, peer_parent, branch_name
        WHERE parent.uuid <> peer_parent.uuid
        """ % {
            "branch_filter": branch_filter,
            "node_kind": self.node_schema.kind,
            "parent_match": parent_match,
            "peer_match": peer_match,
            "peer_parent_match": peer_parent_match,
        }

        self.add_to_query(query)
        self.return_labels = ["active_node.uuid", "parent.uuid", "peer.uuid", "peer_parent.uuid", "branch_name"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()

        for result in self.results:
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("branch_name")),
                    path_type=PathType.RELATIONSHIP_ONE,
                    node_id=str(result.get("peer.uuid")),
                    field_name=self.peer_parent_relationship.name,
                    peer_id=str(result.get("peer_parent.uuid")),
                    kind=self.relationship.peer,
                )
            )

        return grouped_data_paths


class RelationshipPeerParentChecker(ConstraintCheckerInterface):
    query_classes = [RelationshipPeerParentValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "relationship.common_parent.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name and config.SETTINGS.main.schema_strict_mode

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []

        if not request.schema_path.field_name:
            return grouped_data_paths_list

        relationship = request.node_schema.get_relationship(name=request.schema_path.field_name)
        if not relationship.common_parent:
            # Should not happen if schema validation was done properly
            return grouped_data_paths_list

        parent_relationship = next(
            iter(request.node_schema.get_relationships_of_kind(relationship_kinds=[RelationshipKind.PARENT]))
        )
        peer_parent_relationship = request.schema_branch.get(name=relationship.peer, duplicate=False).get_relationship(
            name=relationship.common_parent
        )

        for query_class in self.query_classes:
            query = await query_class.init(
                db=self.db,
                branch=self.branch,
                node_schema=request.node_schema,
                schema_path=request.schema_path,
                relationship=relationship,
                parent_relationship=parent_relationship,
                peer_parent_relationship=peer_parent_relationship,
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())

        return grouped_data_paths_list



from pydantic import BaseModel

from . import internal_schema

INTERNAL_SCHEMA_NODE_KINDS = [node["namespace"] + node["name"] for node in internal_schema["nodes"]]

SUPPORTED_SCHEMA_EXTENSION_TYPE = ["NodeExtensionSchema"]

IGNORE_FOR_NODE = {"id", "state", "filters", "relationships", "attributes"}


class SchemaNamespace(BaseModel):
    name: str
    user_editable: bool



from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema

core_object_template = GenericSchema(
    name="ObjectTemplate",
    namespace="Core",
    include_in_menu=False,
    icon="mdi:pencil-ruler",
    description="Template to create pre-shaped objects.",
    label="Object Templates",
    display_labels=["template_name__value"],
    default_filter="template_name__value",
    uniqueness_constraints=[["template_name__value"]],
    attributes=[
        Attr(
            name="template_name",
            kind="Text",
            description="Unique name identifier for the template",
            optional=False,
            unique=True,
            order_weight=1000,
        )
    ],
)

core_object_component_template = GenericSchema(
    name="ObjectComponentTemplate",
    namespace="Core",
    include_in_menu=False,
    icon="mdi:pencil-ruler",
    description="Component template to create pre-shaped objects.",
    label="Object Component Templates",
    display_labels=["template_name__value"],
    default_filter="template_name__value",
    attributes=[
        Attr(
            name="template_name",
            kind="Text",
            description="Name identifier for the component template",
            optional=False,
            order_weight=1000,
        )
    ],
)



from collections.abc import Sequence
from typing import Any

from ..models import InfrahubObjectParameters
from .data_processor import DataProcessor
from .range_expand_processor import RangeExpandDataProcessor

PROCESSOR_PER_KIND: dict[str, DataProcessor] = {}


class DataProcessorFactory:
    """Factory to create appropriate data processor based on strategy"""

    @classmethod
    def get_processors(cls, kind: str, parameters: InfrahubObjectParameters) -> Sequence[DataProcessor]:
        processors: list[DataProcessor] = []
        if parameters.expand_range:
            processors.append(RangeExpandDataProcessor())
        if kind in PROCESSOR_PER_KIND:
            processors.append(PROCESSOR_PER_KIND[kind])

        return processors

    @classmethod
    async def process_data(
        cls,
        kind: str,
        data: list[dict[str, Any]],
        parameters: InfrahubObjectParameters,
    ) -> list[dict[str, Any]]:
        processors = cls.get_processors(kind=kind, parameters=parameters)
        for processor in processors:
            data = await processor.process_data(data=data)
        return data



from __future__ import annotations

import uuid
from datetime import UTC, datetime, timedelta
from enum import StrEnum
from typing import TYPE_CHECKING, Any

import bcrypt
import jwt
from pydantic import BaseModel

from infrahub import config, models
from infrahub.config import (
    SecurityOAuth2Google,
    SecurityOAuth2Settings,
    SecurityOIDCGoogle,
    SecurityOIDCSettings,
)
from infrahub.core.account import validate_token
from infrahub.core.constants import AccountStatus, InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.protocols import CoreAccount, CoreAccountGroup, CoreGenericAccount
from infrahub.core.registry import registry
from infrahub.exceptions import AuthorizationError, GatewayError, NodeNotFoundError
from infrahub.log import get_logger

if TYPE_CHECKING:
    import httpx

    from infrahub.database import InfrahubDatabase
    from infrahub.services import InfrahubServices

log = get_logger()


class AuthType(StrEnum):
    NONE = "none"
    JWT = "jwt"
    API = "api"


class AccountSession(BaseModel):
    authenticated: bool = True
    account_id: str
    session_id: str | None = None
    auth_type: AuthType

    @property
    def authenticated_by_jwt(self) -> bool:
        return self.auth_type == AuthType.JWT


class SSOStateCache(BaseModel):
    """Cache data stored during OAuth2/OIDC authorization flow.

    This model is used to store state information between the authorization
    request and the token exchange, including PKCE code_verifier when enabled.
    """

    final_url: str
    code_verifier: str | None = None


async def validate_active_account(db: InfrahubDatabase, account_id: str) -> None:
    account = await NodeManager.get_one(db=db, kind=CoreGenericAccount, id=account_id, raise_on_error=True)
    if account.status.value != AccountStatus.ACTIVE.value:
        raise AuthorizationError("This account has been deactivated")


async def authenticate_with_password(
    db: InfrahubDatabase, credentials: models.PasswordCredential, branch: str | None = None
) -> models.UserToken:
    selected_branch = await registry.get_branch(db=db, branch=branch)

    response = await NodeManager.query(
        schema=CoreGenericAccount,
        db=db,
        branch=selected_branch,
        filters={"name__value": credentials.username},
        limit=1,
    )

    if not response:
        raise NodeNotFoundError(
            branch_name=selected_branch.name,
            node_type=InfrahubKind.GENERICACCOUNT,
            identifier=credentials.username,
            message="That login user doesn't exist in the system",
        )

    account = response[0]
    if account.status.value != AccountStatus.ACTIVE.value:
        raise AuthorizationError("This account is not allowed to login")

    password = account.password.value
    valid_credentials = bcrypt.checkpw(credentials.password.encode("UTF-8"), str(password or "").encode("UTF-8"))
    if not valid_credentials:
        raise AuthorizationError("Incorrect password")

    now = datetime.now(tz=UTC)
    refresh_expires = now + timedelta(seconds=config.SETTINGS.security.refresh_token_lifetime)

    session_id = await create_db_refresh_token(db=db, account_id=account.id, expiration=refresh_expires)
    access_token = generate_access_token(account_id=account.id, session_id=session_id)
    refresh_token = generate_refresh_token(account_id=account.id, session_id=session_id, expiration=refresh_expires)

    return models.UserToken(access_token=access_token, refresh_token=refresh_token)


async def create_db_refresh_token(db: InfrahubDatabase, account_id: str, expiration: datetime) -> uuid.UUID:
    obj = await Node.init(db=db, schema=InfrahubKind.REFRESHTOKEN)
    await obj.new(db=db, account=account_id, expiration=expiration.isoformat())
    await obj.save(db=db)
    return uuid.UUID(obj.id)


async def create_fresh_access_token(
    db: InfrahubDatabase, refresh_data: models.RefreshTokenData
) -> models.AccessTokenResponse:
    selected_branch = await registry.get_branch(db=db)

    refresh_token = await NodeManager.get_one(id=str(refresh_data.session_id), db=db)
    if not refresh_token:
        raise AuthorizationError("The provided refresh token has been invalidated in the database")

    account = await NodeManager.get_one(id=refresh_data.account_id, kind=CoreGenericAccount, db=db)
    if not account:
        raise NodeNotFoundError(
            branch_name=selected_branch.name,
            node_type="Account",
            identifier=refresh_data.account_id,
            message="That login user doesn't exist in the system",
        )

    access_token = generate_access_token(account_id=account.id, session_id=refresh_data.session_id)

    return models.AccessTokenResponse(access_token=access_token)


async def signin_sso_account(db: InfrahubDatabase, account_name: str, sso_groups: list[str]) -> models.UserToken:
    account = await NodeManager.get_one_by_default_filter(db=db, id=account_name, kind=InfrahubKind.ACCOUNT)

    if not account:
        account = await Node.init(db=db, schema=InfrahubKind.ACCOUNT)
        await account.new(db=db, name=account_name, account_type="User", password=str(uuid.uuid4()))
        await account.save(db=db)

    if sso_groups:
        infrahub_groups = await NodeManager.query(
            db=db,
            schema=CoreAccountGroup,
            filters={"name__values": sso_groups},
            prefetch_relationships=True,
        )
        for group in infrahub_groups:
            members = await group.members.get_peers(db=db, branch_agnostic=True, peer_type=CoreAccount)
            if account.id not in members:
                await group.members.add(db=db, data=account)
                await group.members.save(db=db)

    now = datetime.now(tz=UTC)
    refresh_expires = now + timedelta(seconds=config.SETTINGS.security.refresh_token_lifetime)
    session_id = await create_db_refresh_token(db=db, account_id=account.id, expiration=refresh_expires)
    access_token = generate_access_token(account_id=account.id, session_id=session_id)
    refresh_token = generate_refresh_token(account_id=account.id, session_id=session_id, expiration=refresh_expires)
    return models.UserToken(access_token=access_token, refresh_token=refresh_token)


def generate_access_token(account_id: str, session_id: uuid.UUID) -> str:
    now = datetime.now(tz=UTC)

    access_expires = now + timedelta(seconds=config.SETTINGS.security.access_token_lifetime)
    access_data = {
        "sub": account_id,
        "iat": now,
        "nbf": now,
        "exp": access_expires,
        "fresh": False,
        "type": "access",
        "session_id": str(session_id),
    }
    access_token = jwt.encode(access_data, config.SETTINGS.security.secret_key, algorithm="HS256")
    return access_token


def generate_refresh_token(account_id: str, session_id: uuid.UUID, expiration: datetime) -> str:
    now = datetime.now(tz=UTC)

    refresh_data = {
        "sub": account_id,
        "iat": now,
        "nbf": now,
        "exp": expiration,
        "fresh": False,
        "type": "refresh",
        "session_id": str(session_id),
    }
    refresh_token = jwt.encode(refresh_data, config.SETTINGS.security.secret_key, algorithm="HS256")
    return refresh_token


async def authentication_token(
    db: InfrahubDatabase, jwt_token: str | None = None, api_key: str | None = None
) -> AccountSession:
    if api_key:
        return await validate_api_key(db=db, token=api_key)
    if jwt_token:
        return await validate_jwt_access_token(token=jwt_token)

    return AccountSession(authenticated=False, account_id="anonymous", auth_type=AuthType.NONE)


async def validate_jwt_access_token(token: str) -> AccountSession:
    try:
        payload = jwt.decode(token, config.SETTINGS.security.secret_key, algorithms=["HS256"])
        account_id = payload["sub"]
        session_id = payload["session_id"]
    except jwt.ExpiredSignatureError:
        raise AuthorizationError("Expired Signature") from None
    except Exception:
        raise AuthorizationError("Invalid token") from None

    if payload["type"] == "access":
        return AccountSession(account_id=account_id, session_id=session_id, auth_type=AuthType.JWT)

    raise AuthorizationError("Invalid token, current token is not an access token")


async def validate_jwt_refresh_token(db: InfrahubDatabase, token: str) -> models.RefreshTokenData:
    try:
        payload = jwt.decode(token, config.SETTINGS.security.secret_key, algorithms=["HS256"])
        account_id = payload["sub"]
        session_id = payload["session_id"]
    except jwt.ExpiredSignatureError:
        raise AuthorizationError("Expired Signature") from None
    except Exception:
        raise AuthorizationError("Invalid token") from None

    await validate_active_account(db=db, account_id=str(account_id))

    if payload["type"] == "refresh":
        return models.RefreshTokenData(account_id=account_id, session_id=session_id)

    raise AuthorizationError("Invalid token, current token is not a refresh token")


async def validate_api_key(db: InfrahubDatabase, token: str) -> AccountSession:
    account_id = await validate_token(token=token, db=db)
    if not account_id:
        raise AuthorizationError("Invalid token")

    await validate_active_account(db=db, account_id=str(account_id))

    return AccountSession(account_id=account_id, auth_type=AuthType.API)


async def invalidate_refresh_token(db: InfrahubDatabase, token_id: str) -> None:
    refresh_token = await NodeManager.get_one(id=token_id, db=db)
    if refresh_token:
        await refresh_token.delete(db=db)


async def get_groups_from_provider(
    provider: SecurityOAuth2Settings | SecurityOIDCSettings, service: InfrahubServices, payload: dict, user_info: dict
) -> list[str]:
    if isinstance(provider, (SecurityOAuth2Google, SecurityOIDCGoogle)):
        # Poor man's workaround to fetch user groups from Google
        if provider.fetch_groups:
            groups_response = await service.http.get(
                f"{provider.cloudidentity_url}?query=member_key_id == '{user_info['email']}'",
                headers={"Authorization": f"{payload.get('token_type')} {payload.get('access_token')}"},
            )
            group_memberships = groups_response.json()
            if "memberships" in group_memberships:
                return [membership["groupKey"]["id"] for membership in group_memberships["memberships"]]

    return []


def safe_get_response_body(response: httpx.Response, raise_error_on_empty_body: bool = True) -> str | dict[str, Any]:
    """Safely extract response body from HTTP response. If the response body cannot be JSON parsed or is empty,
    it raises a GatewayError.

    Args:
        response: The HTTP response object
        raise_error_on_empty_body: Whether to raise an error if the response body is empty

    Returns:
        The response body as JSON dict if possible, otherwise as text

    Raises:
        GatewayError: When the response body cannot be parsed or is empty
    """
    # Try to parse as JSON first
    try:
        return response.json()
    except Exception as json_error:
        try:
            # Try to get as text
            text_body = response.text
            if not text_body.strip() and raise_error_on_empty_body:  # Check for empty or whitespace-only response
                log.error(
                    "Empty response body from authentication provider",
                    url=str(response.url),
                    status_code=response.status_code,
                )
                raise GatewayError(message="Authentication provider returned an empty response") from json_error
        except Exception:
            log.error(
                "Unable to read response body from authentication provider",
                url=str(response.url),
                status_code=response.status_code,
            )
            raise GatewayError(message="Unable to read response from authentication provider") from json_error

    # Here it means we got a text response but not JSON
    return text_body


def extract_auth_error_message(response_body: str | dict[str, Any], base_message: str) -> str:
    """Extract error message from OAuth 2.0/OIDC provider response following RFC 6749.

    Args:
        response_body: The response body from the authentication provider
        base_message: Base error message to use if no specific error is found

    Returns:
        Formatted error message with provider details if available
    """
    if not isinstance(response_body, dict):
        return base_message

    # RFC 6749 standard error response format
    error_description = response_body.get("error_description")
    error_code = response_body.get("error")

    if error_description:
        return f"{base_message}: {error_description}"
    if error_code:
        return f"{base_message}: {error_code}"

    return base_message


def validate_auth_response(response: httpx.Response, provider_type: str = "authentication") -> None:
    """Validate HTTP response from OAuth 2.0/OIDC provider and raise appropriate errors.

    Args:
        response: The HTTP response from the authentication provider
        provider_type: Type of provider for logging (e.g., "OAuth 2.0", "OIDC")

    Raises:
        GatewayError: When the response indicates an error or invalid state
    """
    # If the status code is successful, simply return
    if 200 <= response.status_code <= 299:
        # Verify that we can read the response body safely and it is not empty
        safe_get_response_body(response)
        return

    # Prepare variables with default values for logging
    response_body = safe_get_response_body(response, raise_error_on_empty_body=False)
    log_message: str = f"Unexpected response from {provider_type} provider"
    base_msg: str = "Unexpected response from authentication provider."

    # Handle specific HTTP status codes with appropriate error messages
    match response.status_code:
        case 400:
            log_message = f"Bad request to {provider_type} provider"
            base_msg = "Bad request to authentication provider. Please try again later or contact your administrator."

        case 401:
            log_message = f"Unauthorized request to {provider_type} provider"
            base_msg = (
                "Unauthorized request to authentication provider. Please try again later or contact your administrator."
            )

        case 403:
            log_message = f"Forbidden request to {provider_type} provider"
            base_msg = (
                "Access forbidden by authentication provider. Please try again later or contact your administrator."
            )

        case 404:
            log_message = f"Resource not found for {provider_type} provider"
            base_msg = (
                "Authentication provider endpoint not found. Please try again later or contact your administrator."
            )

        case 429:
            log_message = f"Rate limited by {provider_type} provider"
            base_msg = "Rate limited by authentication provider. Please try again later."

        case status_code if 500 <= status_code <= 599:
            log_message = f"Server error from {provider_type} provider"
            base_msg = "Authentication provider is experiencing server issues. Please try again later or contact your administrator."

    # Print proper log and raise gateway error
    log.error(log_message, url=str(response.url), status_code=response.status_code, body=response_body)
    error_msg = extract_auth_error_message(response_body, base_msg)
    raise GatewayError(message=error_msg)



from __future__ import annotations

import contextvars
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from infrahub.core.node import Node

_creation_context: contextvars.ContextVar[NodeCreationContext | None] = contextvars.ContextVar(
    "node_creation_context", default=None
)


@dataclass
class NodeCreationContext:
    """Collects nodes created as side effects during node creation.

    Uses Python contextvars for async-safe ambient state. Only active when
    explicitly created with a context manager. Recording is a no-op when
    no context is active.
    """

    side_effect_nodes: list[Node] = field(default_factory=list)
    _token: contextvars.Token[NodeCreationContext | None] | None = None

    def record(self, node: Node) -> None:
        self.side_effect_nodes.append(node)

    @classmethod
    def get_current(cls) -> NodeCreationContext | None:
        return _creation_context.get(None)

    @classmethod
    def record_if_active(cls, node: Node) -> None:
        ctx = cls.get_current()
        if ctx is not None:
            ctx.record(node)

    def __enter__(self) -> Self:
        self._token = _creation_context.set(self)
        return self

    def __exit__(self, *args: object) -> None:
        if self._token is not None:
            _creation_context.reset(self._token)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.constants import DiffAction
from infrahub.core.diff.model.path import BranchTrackingId
from infrahub.core.diff.query.merge import (
    DiffMergeMetadataQuery,
    DiffMergeMigratedKindsQuery,
    DiffMergePropertiesQuery,
    DiffMergeQuery,
    DiffMergeRollbackQuery,
)
from infrahub.database import retry_db_transaction
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.diff.model.path import EnrichedDiffRoot
    from infrahub.core.diff.repository.repository import DiffRepository
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase

    from .serializer import DiffMergeSerializer

log = get_logger()


class DiffMerger:
    metadata_batch_size = 500

    def __init__(
        self,
        db: InfrahubDatabase,
        source_branch: Branch,
        destination_branch: Branch,
        diff_repository: DiffRepository,
        serializer: DiffMergeSerializer,
    ) -> None:
        self.source_branch = source_branch
        self.destination_branch = destination_branch
        self.db = db
        self.diff_repository = diff_repository
        self.serializer = serializer
        self._affected_node_uuids: list[str] = []

    async def merge_graph(self, at: Timestamp) -> EnrichedDiffRoot:
        tracking_id = BranchTrackingId(name=self.source_branch.name)
        enriched_diffs = await self.diff_repository.get_roots_metadata(
            diff_branch_names=[self.source_branch.name],
            base_branch_names=[self.destination_branch.name],
            tracking_id=tracking_id,
        )
        latest_diff = None
        for diff in enriched_diffs:
            if latest_diff is None or (diff.to_time > latest_diff.to_time):
                latest_diff = diff
        if latest_diff is None:
            raise RuntimeError(f"Missing diff for branch {self.source_branch.name}")
        log.info(f"Retrieving diff {latest_diff.uuid}")
        enriched_diff = await self.diff_repository.get_one(
            diff_branch_name=self.source_branch.name, diff_id=latest_diff.uuid
        )
        log.info(f"Diff {latest_diff.uuid} retrieved")
        batch_num = 0
        migrated_kinds_id_map = {}
        for n in enriched_diff.nodes:
            if not n.is_node_kind_migration:
                continue
            if n.uuid not in migrated_kinds_id_map or (
                n.uuid in migrated_kinds_id_map and n.action is DiffAction.ADDED
            ):
                # make sure that we use the ADDED db_id if it exists
                # it will not if a node was migrated and then deleted
                migrated_kinds_id_map[n.uuid] = n.identifier.db_id

        async for node_diff_dicts, property_diff_dicts in self.serializer.serialize_diff(diff=enriched_diff):
            if node_diff_dicts:
                log.info(f"Merging batch of nodes #{batch_num}")
                await self._merge_nodes(
                    at=at, node_diff_dicts=node_diff_dicts, migrated_kinds_id_map=migrated_kinds_id_map
                )
            if property_diff_dicts:
                log.info(f"Merging batch of properties #{batch_num}")
                await self._merge_properties(
                    at=at, property_diff_dicts=property_diff_dicts, migrated_kinds_id_map=migrated_kinds_id_map
                )
            log.info(f"Batch #{batch_num} merged")
            batch_num += 1
        migrated_kind_uuids = {n.identifier.uuid for n in enriched_diff.nodes if n.is_node_kind_migration}
        if migrated_kind_uuids:
            migrated_merge_query = await DiffMergeMigratedKindsQuery.init(
                db=self.db,
                branch=self.source_branch,
                at=at,
                target_branch=self.destination_branch,
                migrated_uuids=list(migrated_kind_uuids),
            )
            await migrated_merge_query.execute(db=self.db)

        affected_node_uuids = [n.uuid for n in enriched_diff.nodes]
        self._affected_node_uuids = affected_node_uuids
        if affected_node_uuids:
            for i in range(0, len(affected_node_uuids), self.metadata_batch_size):
                batch_uuids = affected_node_uuids[i : i + self.metadata_batch_size]
                log.info(f"Updating metadata for batch {i // self.metadata_batch_size + 1} ({len(batch_uuids)} nodes)")
                metadata_query = await DiffMergeMetadataQuery.init(
                    db=self.db,
                    branch=self.source_branch,
                    at=at,
                    target_branch=self.destination_branch,
                    node_uuids=batch_uuids,
                )
                await metadata_query.execute(db=self.db)

        # set the branched_from time to the previous microsecond to prevent duplicated
        # relationships on the branch after the merge
        branched_from = at.subtract(microseconds=1)

        self.source_branch.branched_from = branched_from.to_string()
        await self.source_branch.save(db=self.db)
        registry.branch[self.source_branch.name] = self.source_branch
        return enriched_diff

    @retry_db_transaction(name="diff_merge_nodes")
    async def _merge_nodes(
        self, at: Timestamp, node_diff_dicts: list[dict], migrated_kinds_id_map: dict[str, str]
    ) -> None:
        merge_query = await DiffMergeQuery.init(
            db=self.db,
            branch=self.source_branch,
            at=at,
            target_branch=self.destination_branch,
            node_diff_dicts=node_diff_dicts,
            migrated_kinds_id_map=migrated_kinds_id_map,
        )
        await merge_query.execute(db=self.db)

    @retry_db_transaction(name="diff_merge_properties")
    async def _merge_properties(
        self, at: Timestamp, property_diff_dicts: list[dict], migrated_kinds_id_map: dict[str, str]
    ) -> None:
        merge_properties_query = await DiffMergePropertiesQuery.init(
            db=self.db,
            branch=self.source_branch,
            at=at,
            target_branch=self.destination_branch,
            property_diff_dicts=property_diff_dicts,
            migrated_kinds_id_map=migrated_kinds_id_map,
        )
        await merge_properties_query.execute(db=self.db)

    async def rollback(self, at: Timestamp) -> None:
        if not self._affected_node_uuids:
            return
        rollback_query = await DiffMergeRollbackQuery.init(
            db=self.db,
            branch=self.source_branch,
            target_branch=self.destination_branch,
            at=at,
            node_uuids=self._affected_node_uuids,
        )
        await rollback_query.execute(db=self.db)



from __future__ import annotations

import sys
from typing import Any, Self

from pydantic import ConfigDict, Field, model_validator

from infrahub import config
from infrahub.core.constants.schema import UpdateSupport
from infrahub.core.models import HashableModel
from infrahub.exceptions import ValidationError


def get_attribute_parameters_class_for_kind(kind: str) -> type[AttributeParameters]:
    param_classes: dict[str, type[AttributeParameters]] = {
        "NumberPool": NumberPoolParameters,
        "Text": TextAttributeParameters,
        "TextArea": TextAttributeParameters,
        "List": ListAttributeParameters,
        "Number": NumberAttributeParameters,
    }
    return param_classes.get(kind, AttributeParameters)


class AttributeParameters(HashableModel):
    model_config = ConfigDict(extra="forbid")

    @classmethod
    def convert_from(cls, source: AttributeParameters) -> Self:
        """Convert from another AttributeParameters subclass.

        Args:
            source: The source AttributeParameters instance to convert from

        Returns:
            A new instance of the target class with compatible fields populated
        """
        source_data = source.model_dump()
        return cls.convert_from_dict(source_data=source_data)

    @classmethod
    def convert_from_dict(cls, source_data: dict[str, Any]) -> Self:
        """Convert from a dictionary to the target class.

        Args:
            source_data: The source dictionary to convert from

        Returns:
            A new instance of the target class with compatible fields populated
        """
        target_fields = set(cls.model_fields.keys())
        filtered_data = {k: v for k, v in source_data.items() if k in target_fields}
        return cls(**filtered_data)


class ListAttributeParameters(AttributeParameters):
    """Parameters for List attributes supporting regex validation on list items."""

    regex: str | None = Field(
        default=None,
        description="Regular expression that each list item value must match if defined",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )


class TextAttributeParameters(AttributeParameters):
    regex: str | None = Field(
        default=None,
        description="Regular expression that attribute value must match if defined",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )
    min_length: int | None = Field(
        default=None,
        description="Set a minimum number of characters allowed.",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )
    max_length: int | None = Field(
        default=None,
        description="Set a maximum number of characters allowed.",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )

    @model_validator(mode="after")
    def validate_min_max(self) -> Self:
        if (
            config.SETTINGS.initialized
            and config.SETTINGS.main.schema_strict_mode
            and self.min_length is not None
            and self.max_length is not None
        ):
            if self.min_length > self.max_length:
                raise ValueError(
                    "`max_length` can't be less than `min_length` when the schema is configured with strict mode"
                )

        return self


class NumberAttributeParameters(AttributeParameters):
    min_value: int | None = Field(
        default=None,
        description="Set a minimum value allowed.",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )
    max_value: int | None = Field(
        default=None,
        description="Set a maximum value allowed.",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )
    excluded_values: str | None = Field(
        default=None,
        description="List of values or range of values not allowed for the attribute, format is: '100,150-200,280,300-400'",
        pattern=r"^(\d+(?:-\d+)?)(?:,\d+(?:-\d+)?)*$",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )

    @model_validator(mode="after")
    def validate_ranges(self) -> Self:
        ranges = self.get_excluded_ranges()
        for i, (start_range_1, end_range_1) in enumerate(ranges):
            if start_range_1 > end_range_1:
                raise ValueError("`start_range` can't be less than `end_range`")

            # Check for overlapping ranges
            for start_range_2, end_range_2 in ranges[i + 1 :]:
                if not (end_range_1 < start_range_2 or start_range_1 > end_range_2):
                    raise ValueError("Excluded ranges cannot overlap")

        return self

    @model_validator(mode="after")
    def validate_min_max(self) -> Self:
        if (
            config.SETTINGS.initialized
            and config.SETTINGS.main.schema_strict_mode
            and self.min_value is not None
            and self.max_value is not None
        ):
            if self.min_value > self.max_value:
                raise ValueError(
                    "`max_value` can't be less than `min_value` when the schema is configured with strict mode"
                )

        return self

    def get_excluded_single_values(self) -> list[int]:
        if not self.excluded_values:
            return []

        results = [int(value) for value in self.excluded_values.split(",") if "-" not in value]
        return results

    def get_excluded_ranges(self) -> list[tuple[int, int]]:
        if not self.excluded_values:
            return []

        ranges = []
        for value in self.excluded_values.split(","):
            if "-" in value:
                start, end = map(int, value.split("-"))
                ranges.append((start, end))

        return ranges

    def is_valid_value(self, value: int) -> bool:
        try:
            self.check_valid_value(value=value, name="UNUSED")
        except ValidationError:
            return False
        return True

    def check_valid_value(self, value: int, name: str) -> None:
        if self.min_value is not None and value < self.min_value:
            raise ValidationError({name: f"{value} is lower than the minimum allowed value {self.min_value!r}"})
        if self.max_value is not None and value > self.max_value:
            raise ValidationError({name: f"{value} is higher than the maximum allowed value {self.max_value!r}"})
        if value in self.get_excluded_single_values():
            raise ValidationError({name: f"{value} is in the excluded values"})
        for start, end in self.get_excluded_ranges():
            if start <= value <= end:
                raise ValidationError({name: f"{value} is in an the excluded range {start}-{end}"})


class NumberPoolParameters(AttributeParameters):
    end_range: int = Field(
        default=sys.maxsize,
        description="End range for numbers for the associated NumberPool",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )
    start_range: int = Field(
        default=1,
        description="Start range for numbers for the associated NumberPool",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )
    number_pool_id: str | None = Field(
        default=None,
        description="The ID of the numberpool associated with this attribute. Only set after the number pool has been provisioned.",
        json_schema_extra={"update": UpdateSupport.NOT_SUPPORTED.value},
    )

    @model_validator(mode="after")
    def validate_ranges(self) -> Self:
        if self.start_range > self.end_range:
            raise ValueError("`start_range` can't be less than `end_range`")
        return self

    def get_pool_size(self) -> int:
        """
        Returns the size of the pool based on the defined ranges.
        """
        return self.end_range - self.start_range + 1



from infrahub.database import InfrahubDatabase

from .menu import default_menu
from .repository import MenuRepository


async def create_default_menu(db: InfrahubDatabase) -> None:
    repository = MenuRepository(db=db)
    await repository.create_menu(menu=default_menu)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.migrations.graph.m041_deleted_dup_edges import DeleteDuplicatedRelationshipEdges
from infrahub.core.migrations.shared import ArbitraryMigration, MigrationInput, MigrationResult, get_migration_console
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class UndeleteRelationshipProperties(Query):
    """
    Find Relationship vertices that are missing IS_VISIBLE and/or IS_PROTECTED edges linking them to Boolean vertices

    Use the existing IS_RELATED edges to determine when the IS_VISIBLE/IS_PROTECTED edges should exist on each branch
    and add the missing edges

    Sets IS_VISIBLE to TRUE and IS_PROTECTED to FALSE
    """

    name = "undelete_relationship_properties"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
// --------------
// find all Relationships that are missing IS_VISIBLE/IS_PROTECTED edges
// --------------
MATCH (rel:Relationship)-[:IS_RELATED]-()
WHERE NOT exists((rel)-[:IS_VISIBLE]->())
OR NOT exists((rel)-[:IS_PROTECTED]->())
WITH DISTINCT rel
WITH rel, exists((rel)-[:IS_VISIBLE]->()) AS has_visible, exists((rel)-[:IS_PROTECTED]->()) AS has_protected
WHERE has_visible = FALSE OR has_protected = FALSE

// --------------
// one row for each branch the Relationship has changes on
// --------------
CALL (rel) {
    MATCH (rel)-[e:IS_RELATED]-()
    RETURN DISTINCT e.branch AS branch
}

// --------------
// get the earliest from edge on this branch, if any
// --------------
CALL (rel, branch) {
    OPTIONAL MATCH (rel)-[e:IS_RELATED {status: "active", branch: branch}]-()
    // ignore schema kind/inheritance migration edges
    WHERE NOT exists((rel)-[:IS_RELATED {status: "deleted", branch: branch, from: e.from}]-())
    RETURN e AS earliest_from_edge
    ORDER BY e.from ASC
    LIMIT 1
}

// --------------
// get the latest to time on this branch, if any
// --------------
CALL (rel, branch) {
    OPTIONAL MATCH (rel)-[e:IS_RELATED {status: "active", branch: branch}]-()
    // ignore schema kind/inheritance migration edges
    WHERE NOT exists((rel)-[:IS_RELATED {status: "active", branch: branch, from: e.to}]-())
    RETURN e.to AS latest_to_time
    ORDER BY e.to DESC
    LIMIT 1
}

// --------------
// get the latest deleted edge on this branch, if any
// --------------
CALL (rel, branch) {
    OPTIONAL MATCH (rel)-[deleted_e:IS_RELATED {status: "deleted", branch: branch}]-()
    // ignore schema kind/inheritance migration edges
    WHERE NOT exists((rel)-[:IS_RELATED {status: "active", branch: branch, from: deleted_e.from}]-())
    RETURN deleted_e AS latest_deleted_edge
    ORDER BY deleted_e.from DESC
    LIMIT 1
}

// --------------
// add active IS_VISIBLE edge on the branch, if necessary
// --------------
CALL (rel, earliest_from_edge, latest_to_time, has_visible) {
    WITH *, has_visible
    WHERE has_visible = FALSE
    AND earliest_from_edge IS NOT NULL
    MERGE (bool:Boolean {value: TRUE})
    CREATE (rel)-[new_edge:IS_VISIBLE]->(bool)
    SET new_edge = properties(earliest_from_edge)
    SET new_edge.to = latest_to_time
}
// --------------
// add deleted IS_VISIBLE edge on the branch, if necessary
// --------------
CALL (rel, latest_deleted_edge, has_visible) {
    WITH *, has_visible
    WHERE has_visible = FALSE
    AND latest_deleted_edge IS NOT NULL
    MERGE (bool:Boolean {value: TRUE})
    CREATE (rel)-[new_edge:IS_VISIBLE]->(bool)
    SET new_edge = properties(latest_deleted_edge)
}
// --------------
// add active IS_PROTECTED edge on the branch, if necessary
// --------------
CALL (rel, earliest_from_edge, latest_to_time, has_protected) {
    WITH *, has_protected
    WHERE has_protected = FALSE
    AND earliest_from_edge IS NOT NULL
    MERGE (bool:Boolean {value: FALSE})
    CREATE (rel)-[new_edge:IS_PROTECTED]->(bool)
    SET new_edge = properties(earliest_from_edge)
    SET new_edge.to = latest_to_time
}
// --------------
// add deleted IS_PROTECTED edge on the branch, if necessary
// --------------
CALL (rel, latest_deleted_edge, has_protected) {
    WITH *, has_protected
    WHERE has_protected = FALSE
    AND latest_deleted_edge IS NOT NULL
    MERGE (bool:Boolean {value: FALSE})
    CREATE (rel)-[new_edge:IS_PROTECTED]->(bool)
    SET new_edge = properties(latest_deleted_edge)
}
        """
        self.add_to_query(query)


class Migration048(ArbitraryMigration):
    """
    Fix Relationship vertices that are missing IS_VISIBLE and/or IS_PROTECTED edges.

    This can happen due to a bug in Migration041 that deleted these edges incorrectly.
    """

    name: str = "048_undelete_rel_props"
    minimum_version: int = 47

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        console = get_migration_console()

        console.log("Deleting duplicate edges for all Relationships", end="...")
        delete_duplicate_edges_query = await DeleteDuplicatedRelationshipEdges.init(
            db=db, migrated_kind_nodes_only=False, at=at
        )
        await delete_duplicate_edges_query.execute(db=db)
        console.log("done")

        console.log("Undeleting Relationship properties", end="...")
        undelete_rel_props_query = await UndeleteRelationshipProperties.init(db=db, at=at)
        await undelete_rel_props_query.execute(db=db)
        console.log("done")

        return MigrationResult()



from __future__ import annotations

from prefect import flow
from prefect.logging import get_run_logger

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.branch.models import Branch
from infrahub.core.registry import registry
from infrahub.message_bus.messages.refresh_registry_branches import RefreshRegistryBranches
from infrahub.pools.schema_number_pool_synchronizer import SchemaNumberPoolSynchronizer
from infrahub.pools.schema_number_pool_upserter import SchemaNumberPoolUpserter
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow


@flow(
    name="validate-schema-number-pools",
    flow_run_name="Validate schema number pools",
)
async def validate_schema_number_pools(
    branch_name: str,  # noqa: ARG001
    context: InfrahubContext,
    service: InfrahubServices,
) -> set[str]:
    log = get_run_logger()
    synchronizer = SchemaNumberPoolSynchronizer(
        db=service.database,
        schema_manager=registry.schema,
        upserter=SchemaNumberPoolUpserter(db=service.database, schema_manager=registry.schema),
        log=log,
    )
    updated_branches = await synchronizer.run(user_id=context.account.account_id)

    if updated_branches:
        for updated_branch_name in updated_branches:
            branch = await Branch.get_by_name(db=service.database, name=updated_branch_name)
            branch.update_schema_hash()
            await branch.save(db=service.database, user_id=context.account.account_id)
            log.info(f"Updated schema hash for branch {updated_branch_name} after number pool synchronization")

        await service.component.refresh_schema_hash(branches=list(updated_branches))
        await service.message_bus.send(RefreshRegistryBranches())

    return updated_branches



import re
from collections import defaultdict
from typing import Any

from infrahub_sdk import InfrahubClient
from infrahub_sdk.node import RelationshipManager
from infrahub_sdk.protocols import (
    CoreArtifactDefinition,
    CoreCheckDefinition,
    CoreGroup,
    CoreReadOnlyRepository,
    CoreRepository,
)
from infrahub_sdk.types import Order

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.order import OrderModel
from infrahub.database import InfrahubDatabase
from infrahub.generators.models import ProposedChangeGeneratorDefinition

from .. import config
from .models import RepositoryBranchInfo, RepositoryData


async def get_repositories_commit_per_branch(
    db: InfrahubDatabase,
    kind: str = InfrahubKind.GENERICREPOSITORY,
) -> dict[str, RepositoryData]:
    """Get a list of all repositories and their commit on each branches.

    This method is similar to 'get_list_repositories' method in the Python SDK.

    NOTE: At some point, we should refactor this function to use a single Database query instead of one per branch
    """

    repositories: dict[str, RepositoryData] = {}

    for branch in list(registry.branch.values()):
        repos: list[CoreRepository | CoreReadOnlyRepository] = await NodeManager.query(
            db=db,
            branch=branch,
            fields={"id": None, "name": None, "commit": None, "internal_status": None, "location": None, "ref": None},
            schema=kind,
            order=OrderModel(disable=True),
        )

        for repository in repos:
            repo_name = repository.name.value
            if repo_name not in repositories:
                repositories[repo_name] = RepositoryData(
                    repository_id=repository.get_id(),
                    repository_name=repo_name,
                    repository=repository,
                    branches={},
                )

            repositories[repo_name].branches[branch.name] = repository.commit.value
            repositories[repo_name].branch_info[branch.name] = RepositoryBranchInfo(
                internal_status=repository.internal_status.value
            )

    return repositories


def _collect_parameter_first_segments(params: Any) -> set[str]:
    segments: set[str] = set()

    def _walk(value: Any) -> None:
        if isinstance(value, str):
            segment = value.split("__", 1)[0]
            if segment:
                segments.add(segment)
        elif isinstance(value, dict):
            for nested in value.values():
                _walk(nested)
        elif isinstance(value, (list, tuple, set)):
            for nested in value:
                _walk(nested)

    _walk(params)
    return segments


async def _prefetch_group_member_nodes(
    client: InfrahubClient,
    members: RelationshipManager,
    branch: str,
    required_fields: set[str],
) -> None:
    ids_per_kind: dict[str, set[str]] = defaultdict(set)
    for peer in members.peers:
        if peer.id and peer.typename:
            ids_per_kind[peer.typename].add(peer.id)

    if not ids_per_kind:
        return

    batch = await client.create_batch()

    for kind, ids in ids_per_kind.items():
        schema = await client.schema.get(kind=kind, branch=branch)

        # FIXME: https://github.com/opsmill/infrahub-sdk-python/pull/205
        valid_fields = set(schema.attribute_names) | set(schema.relationship_names)
        keep_relationships = set(schema.relationship_names) & required_fields
        cleaned_fields = valid_fields - required_fields

        kwargs: dict[str, Any] = {
            "kind": kind,
            "ids": list(ids),
            "branch": branch,
            "exclude": list(cleaned_fields),
            "populate_store": True,
            "order": Order(disable=True),
        }

        if keep_relationships:
            kwargs["include"] = list(keep_relationships)

        batch.add(task=client.filters, **kwargs)

    async for _ in batch.execute():
        pass


async def _fetch_definition_targets(
    client: InfrahubClient,
    branch: str,
    group_id: str,
    parameters: Any,
) -> CoreGroup:
    group = await client.get(
        kind=CoreGroup,
        id=group_id,
        branch=branch,
        include=["members"],
    )

    parameter_fields = _collect_parameter_first_segments(parameters)
    await _prefetch_group_member_nodes(
        client=client,
        members=group.members,
        branch=branch,
        required_fields=parameter_fields,
    )

    return group


async def fetch_artifact_definition_targets(
    client: InfrahubClient,
    branch: str,
    definition: CoreArtifactDefinition,
) -> CoreGroup:
    return await _fetch_definition_targets(
        client=client, branch=branch, group_id=definition.targets.id, parameters=definition.parameters.value
    )


async def fetch_check_definition_targets(
    client: InfrahubClient,
    branch: str,
    definition: CoreCheckDefinition,
) -> CoreGroup:
    return await _fetch_definition_targets(
        client=client, branch=branch, group_id=definition.targets.id, parameters=definition.parameters.value
    )


async def fetch_proposed_change_generator_definition_targets(
    client: InfrahubClient,
    branch: str,
    definition: ProposedChangeGeneratorDefinition,
) -> CoreGroup:
    return await _fetch_definition_targets(
        client=client, branch=branch, group_id=definition.group_id, parameters=definition.parameters
    )


def branch_name_in_import_sync_branches(branch_short_name: str) -> bool:
    for branch_filter in config.SETTINGS.git.import_sync_branch_names:
        if re.fullmatch(branch_filter, branch_short_name) or branch_filter == branch_short_name:
            return True
    return False



import logging
import sys
from datetime import datetime, timezone
from typing import TYPE_CHECKING

import typer
from rich.console import Console
from rich.table import Table

from ..async_typer import AsyncTyper
from ..branch import BranchData
from ..diff import DiffTreeData
from ..protocols import CoreProposedChange
from ..utils import calculate_time_diff, decode_json
from .client import initialize_client
from .parameters import CONFIG_PARAM
from .utils import catch_exception

if TYPE_CHECKING:
    from ..client import InfrahubClient

app = AsyncTyper()
console = Console()


DEFAULT_CONFIG_FILE = "infrahubctl.toml"
ENVVAR_CONFIG_FILE = "INFRAHUBCTL_CONFIG"


def format_timestamp(timestamp: str) -> str:
    """Format ISO timestamp to 'YYYY-MM-DD HH:MM:SS'.
    Args:
        timestamp (str): ISO fromatted timestamp

    Returns:
        (str): the datetime as string formatted as 'YYYY-MM-DD HH:MM:SS'

    Raises:
        Any execptions returned from formatting the timestamp are propogated to the caller
    """
    dt = datetime.fromisoformat(timestamp.replace("Z", "+00:00"))
    return dt.strftime("%Y-%m-%d %H:%M:%S")


async def check_git_files_changed(client: "InfrahubClient", branch: str) -> bool:
    """Check if there are any Git file changes in a branch.

    Args:
        client: Infrahub client instance
        branch: Branch name to check

    Returns:
        True if files have changed, False otherwise

    Raises:
        Any exceptions from the API call are propagated to the caller
    """
    url = f"{client.address}/api/diff/files?branch={branch}"
    resp = await client._get(url=url, timeout=client.default_timeout)
    resp.raise_for_status()
    data = decode_json(response=resp)

    # Check if any repository has files
    if branch in data:
        for repo_data in data[branch].values():
            if isinstance(repo_data, dict) and "files" in repo_data and len(repo_data["files"]) > 0:
                return True

    return False


def generate_branch_report_table(
    branch: BranchData, diff_tree: DiffTreeData | None, git_files_changed: bool | None
) -> Table:
    branch_table = Table(show_header=False, box=None)
    branch_table.add_column(justify="left")
    branch_table.add_column(justify="right")

    branch_table.add_row("Created at", format_timestamp(branch.branched_from))

    status_value = branch.status.value if hasattr(branch.status, "value") else str(branch.status)
    branch_table.add_row("Status", status_value)

    branch_table.add_row("Synced with Git", "Yes" if branch.sync_with_git else "No")

    if git_files_changed is not None:
        branch_table.add_row("Git files changed", "Yes" if git_files_changed else "No")
    else:
        branch_table.add_row("Git files changed", "N/A")

    branch_table.add_row("Has schema changes", "Yes" if branch.has_schema_changes else "No")

    if diff_tree:
        branch_table.add_row("Diff last updated", format_timestamp(diff_tree["to_time"]))
        branch_table.add_row("Amount of additions", str(diff_tree["num_added"]))
        branch_table.add_row("Amount of deletions", str(diff_tree["num_removed"]))
        branch_table.add_row("Amount of updates", str(diff_tree["num_updated"]))
        branch_table.add_row("Amount of conflicts", str(diff_tree["num_conflicts"]))
    else:
        branch_table.add_row("Diff last updated", "No diff available")
        branch_table.add_row("Amount of additions", "-")
        branch_table.add_row("Amount of deletions", "-")
        branch_table.add_row("Amount of updates", "-")
        branch_table.add_row("Amount of conflicts", "-")

    return branch_table


def generate_proposed_change_tables(proposed_changes: list[CoreProposedChange]) -> list[Table]:
    proposed_change_tables: list[Table] = []

    for pc in proposed_changes:
        metadata = pc.get_node_metadata()
        created_by = metadata.created_by.display_label if metadata and metadata.created_by else "-"
        created_at = format_timestamp(metadata.created_at) if metadata and metadata.created_at else "-"

        # Create proposal table
        proposed_change_table = Table(show_header=False, box=None)
        proposed_change_table.add_column(justify="left")
        proposed_change_table.add_column(justify="right")

        # Extract data from node
        proposed_change_table.add_row("Name", pc.name.value)
        proposed_change_table.add_row("State", str(pc.state.value))
        proposed_change_table.add_row("Is draft", "Yes" if pc.is_draft.value else "No")
        proposed_change_table.add_row("Created by", created_by)
        proposed_change_table.add_row("Created at", created_at)
        proposed_change_table.add_row("Approvals", str(len(pc.approved_by.peers)))
        proposed_change_table.add_row("Rejections", str(len(pc.rejected_by.peers)))

        proposed_change_tables.append(proposed_change_table)

    return proposed_change_tables


@app.callback()
def callback() -> None:
    """
    Manage the branches in a remote Infrahub instance.

    List, create, merge, rebase ..
    """


@app.command("list")
@catch_exception(console=console)
async def list_branch(_: str = CONFIG_PARAM) -> None:
    """List all existing branches."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    branches = await client.branch.all()

    table = Table(title="List of all branches")

    table.add_column("Name", justify="right", style="cyan", no_wrap=True)
    table.add_column("Description")
    table.add_column("Origin Branch")
    table.add_column("Branched From")
    table.add_column("Sync with Git")
    table.add_column("Has Schema Changes")
    table.add_column("Is Default")
    table.add_column("Status")

    # identify the default branch and always print it first
    default_branch = next(branch for branch in branches.values() if branch.is_default)
    table.add_row(
        default_branch.name,
        default_branch.description or " - ",
        default_branch.origin_branch,
        f"{default_branch.branched_from} ({calculate_time_diff(default_branch.branched_from)})",
        "[green]True" if default_branch.sync_with_git else "[#FF7F50]False",
        "[green]True" if default_branch.has_schema_changes else "[#FF7F50]False",
        "[green]True" if default_branch.is_default else "[#FF7F50]False",
        default_branch.status,
    )

    for branch in branches.values():
        if branch.is_default:
            continue

        table.add_row(
            branch.name,
            branch.description or " - ",
            branch.origin_branch,
            f"{branch.branched_from} ({calculate_time_diff(branch.branched_from)})",
            "[green]True" if branch.sync_with_git else "[#FF7F50]False",
            "[green]True" if default_branch.has_schema_changes else "[#FF7F50]False",
            "[green]True" if branch.is_default else "[#FF7F50]False",
            branch.status,
        )

    console.print(table)


@app.command()
@catch_exception(console=console)
async def create(
    branch_name: str = typer.Argument(..., help="Name of the branch to create"),
    description: str = typer.Option(default="", help="Description of the branch"),
    sync_with_git: bool = typer.Option(
        False, help="Extend the branch to Git and have Infrahub create the branch in connected repositories."
    ),
    isolated: bool = typer.Option(True, hidden=True, help="Set the branch to isolated mode (deprecated)"),  # noqa: ARG001
    _: str = CONFIG_PARAM,
) -> None:
    """Create a new branch."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    branch = await client.branch.create(branch_name=branch_name, description=description, sync_with_git=sync_with_git)
    console.print(f"Branch {branch_name!r} created successfully ({branch.id}).")


@app.command()
@catch_exception(console=console)
async def delete(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Delete a branch."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    await client.branch.delete(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' deleted successfully.")


@app.command()
@catch_exception(console=console)
async def rebase(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Rebase a Branch with main."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    await client.branch.rebase(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' rebased successfully.")


@app.command()
@catch_exception(console=console)
async def merge(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Merge a Branch with main."""

    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)

    client = initialize_client()
    await client.branch.merge(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' merged successfully.")


@app.command()
@catch_exception(console=console)
async def validate(branch_name: str, _: str = CONFIG_PARAM) -> None:
    """Validate if a branch has some conflict and is passing all the tests (NOT IMPLEMENTED YET)."""

    client = initialize_client()
    await client.branch.validate(branch_name=branch_name)
    console.print(f"Branch '{branch_name}' is valid.")


@app.command()
@catch_exception(console=console)
async def report(
    branch_name: str = typer.Argument(..., help="Branch name to generate report for"),
    update_diff: bool = typer.Option(False, "--update-diff", help="Update diff before generating report"),
    _: str = CONFIG_PARAM,
) -> None:
    """Generate branch cleanup status report."""

    client = initialize_client()

    # Fetch branch metadata first (needed for diff creation)
    branch = await client.branch.get(branch_name=branch_name)

    if branch.is_default:
        console.print("[red]Cannot create a report for the default branch!")
        sys.exit(1)

    # Update diff if requested
    if update_diff:
        console.print("Updating diff...")
        # Create diff from branch creation to now
        from_time = datetime.fromisoformat(branch.branched_from.replace("Z", "+00:00"))
        to_time = datetime.now(timezone.utc)
        await client.create_diff(
            branch=branch_name,
            name=f"report-{branch_name}",
            from_time=from_time,
            to_time=to_time,
        )
        console.print("Diff updated\n")

    diff_tree = await client.get_diff_tree(branch=branch_name)

    git_files_changed = await check_git_files_changed(client, branch=branch_name)

    proposed_changes = await client.filters(
        kind=CoreProposedChange,  # type: ignore[type-abstract]
        source_branch__value=branch_name,
        prefetch_relationships=True,
        property=True,
        include_metadata=True,
    )

    branch_table = generate_branch_report_table(branch=branch, diff_tree=diff_tree, git_files_changed=git_files_changed)
    proposed_change_tables = generate_proposed_change_tables(proposed_changes=proposed_changes)

    console.print()
    console.print(f"[bold]Branch: {branch_name}[/bold]")

    console.print(branch_table)
    console.print()

    if not proposed_changes:
        console.print("No proposed changes for this branch")
        console.print()

    for proposed_change, proposed_change_table in zip(proposed_changes, proposed_change_tables, strict=True):
        console.print(f"Proposed change: {proposed_change.name.value}")
        console.print(proposed_change_table)
        console.print()



from __future__ import annotations

from typing import Any

from .renderers import render_input_block, render_query_block, render_variables_to_string


class BaseGraphQLQuery:
    query_type: str = "not-defined"
    indentation: int = 4

    def __init__(self, query: dict, variables: dict | None = None, name: str | None = None) -> None:
        self.query = query
        self.variables = variables
        self.name = name or ""

    def render_first_line(self) -> str:
        first_line = self.query_type

        if self.name:
            first_line += " " + self.name

        if self.variables:
            first_line += f" ({render_variables_to_string(self.variables)})"

        first_line += " {"

        return first_line


class Query(BaseGraphQLQuery):
    query_type = "query"

    def render(self, convert_enum: bool = False) -> str:
        lines = [self.render_first_line()]
        lines.extend(
            render_query_block(
                data=self.query, indentation=self.indentation, offset=self.indentation, convert_enum=convert_enum
            )
        )
        lines.append("}")

        return "\n" + "\n".join(lines) + "\n"


class Mutation(BaseGraphQLQuery):
    query_type = "mutation"

    def __init__(self, *args: Any, mutation: str, input_data: dict, **kwargs: Any) -> None:
        self.input_data = input_data
        self.mutation = mutation
        super().__init__(*args, **kwargs)

    def render(self, convert_enum: bool = False) -> str:
        lines = [self.render_first_line()]
        lines.append(" " * self.indentation + f"{self.mutation}(")
        lines.extend(
            render_input_block(
                data=self.input_data,
                indentation=self.indentation,
                offset=self.indentation * 2,
                convert_enum=convert_enum,
            )
        )
        lines.append(" " * self.indentation + "){")
        lines.extend(
            render_query_block(
                data=self.query,
                indentation=self.indentation,
                offset=self.indentation * 2,
                convert_enum=convert_enum,
            )
        )
        lines.extend((" " * self.indentation + "}", "}"))

        return "\n" + "\n".join(lines) + "\n"



from infrahub.core.constants import (
    BranchSupportType,
    GeneratorInstanceStatus,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_generator_definition = NodeSchema(
    name="GeneratorDefinition",
    namespace="Core",
    description="Defines a generator that creates or updates objects based on a GraphQL query",
    include_in_menu=False,
    icon="mdi:state-machine",
    label="Generator Definition",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.AWARE,
    uniqueness_constraints=[["name__value"]],
    generate_profile=False,
    inherit_from=[InfrahubKind.TASKTARGET],
    documentation="/topics/generator",
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="description", kind="Text", optional=True),
        Attr(name="parameters", kind="JSON", description="GraphQL query parameters for the generator"),
        Attr(name="file_path", kind="Text", description="Path to the Python file containing the generator class"),
        Attr(name="class_name", kind="Text", description="Name of the Python class implementing the generator"),
        Attr(
            name="convert_query_response",
            kind="Boolean",
            description="Whether to convert the GraphQL response to SDK objects",
            optional=True,
            default_value=False,
        ),
        Attr(
            name="execute_in_proposed_change",
            kind="Boolean",
            description="Whether to run this generator in proposed changes",
            optional=True,
            default_value=True,
        ),
        Attr(
            name="execute_after_merge",
            kind="Boolean",
            description="Whether to run this generator after a merge",
            optional=True,
            default_value=True,
        ),
    ],
    relationships=[
        Rel(
            name="query",
            peer=InfrahubKind.GRAPHQLQUERY,
            identifier="generator_definition__graphql_query",
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            optional=False,
        ),
        Rel(
            name="repository",
            peer=InfrahubKind.GENERICREPOSITORY,
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            identifier="generator_definition__repository",
            optional=False,
        ),
        Rel(
            name="targets",
            peer=InfrahubKind.GENERICGROUP,
            kind=RelKind.ATTRIBUTE,
            identifier="generator_definition___group",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
    ],
)

core_generator_instance = NodeSchema(
    name="GeneratorInstance",
    namespace="Core",
    description="An instance of a generator execution for a specific object",
    label="Generator Instance",
    include_in_menu=False,
    icon="mdi:file-document-outline",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.LOCAL,
    generate_profile=False,
    inherit_from=[InfrahubKind.TASKTARGET],
    documentation="/topics/generator",
    attributes=[
        Attr(name="name", kind="Text"),
        Attr(name="status", kind="Text", enum=GeneratorInstanceStatus.available_types()),
    ],
    relationships=[
        Rel(
            name="object",
            peer=InfrahubKind.NODE,
            kind=RelKind.ATTRIBUTE,
            identifier="generator__node",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
        Rel(
            name="definition",
            peer=InfrahubKind.GENERATORDEFINITION,
            kind=RelKind.ATTRIBUTE,
            identifier="generator__generator_definition",
            cardinality=Cardinality.ONE,
            optional=False,
        ),
    ],
)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from infrahub.core.node import Node
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import MainSchemaTypes


class PoolAllocator(ABC):
    """Interface for pool allocation strategies."""

    @abstractmethod
    async def allocate_for_attribute(
        self,
        pool_relationship: RelationshipManager,
        target_schema: MainSchemaTypes,
        attribute_name: str,
        identifier: str,
    ) -> Any | None:
        """Allocate from a _from_resource_pool relationship for an attribute (e.g., Number → CoreNumberPool)."""

    @abstractmethod
    async def allocate_for_relationship(self, pool_relationship: RelationshipManager, identifier: str) -> Node | None:
        """Allocate from resource pool for a _from_resource_pool relationship."""



from .combiner import DiffCombiner
from .model.path import (
    EnrichedDiffAttribute,
    EnrichedDiffNode,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
    EnrichedDiffSingleRelationship,
)


class DiffConflictTransferer:
    def __init__(self, diff_combiner: DiffCombiner) -> None:
        self.diff_combiner = diff_combiner

    async def transfer(self, earlier: EnrichedDiffRoot, later: EnrichedDiffRoot) -> None:
        earlier_node_map = {n.uuid: n for n in earlier.nodes}
        later_node_map = {n.uuid: n for n in later.nodes}
        common_node_uuids = set(earlier_node_map.keys()) & set(later_node_map.keys())
        for node_uuid in common_node_uuids:
            earlier_node = earlier_node_map[node_uuid]
            later_node = later_node_map[node_uuid]
            self._transfer_node_conflicts(earlier=earlier_node, later=later_node)

    def _transfer_node_conflicts(self, earlier: EnrichedDiffNode, later: EnrichedDiffNode) -> None:
        if earlier.action is later.action:
            later.conflict = self.diff_combiner.combine_conflicts(earlier=earlier.conflict, later=later.conflict)

        earlier_attribute_map = {a.name: a for a in earlier.attributes}
        later_attribute_map = {a.name: a for a in later.attributes}
        common_attribute_names = set(earlier_attribute_map.keys()) & set(later_attribute_map.keys())
        for attr_name in common_attribute_names:
            earlier_attribute = earlier_attribute_map[attr_name]
            later_attribute = later_attribute_map[attr_name]
            self._transfer_property_conflicts(earlier=earlier_attribute, later=later_attribute)
        earlier_relationship_map = {r.name: r for r in earlier.relationships}
        later_relationship_map = {r.name: r for r in later.relationships}
        common_relationship_names = set(earlier_relationship_map.keys()) & set(later_relationship_map.keys())
        for relationship_name in common_relationship_names:
            earlier_relationship = earlier_relationship_map[relationship_name]
            later_relationship = later_relationship_map[relationship_name]
            self._transfer_relationship_conflicts(
                earlier=earlier_relationship,
                later=later_relationship,
            )

    def _transfer_property_conflicts(
        self,
        earlier: EnrichedDiffAttribute | EnrichedDiffSingleRelationship,
        later: EnrichedDiffAttribute | EnrichedDiffSingleRelationship,
    ) -> None:
        earlier_property_map = {p.property_type: p for p in earlier.properties}
        later_property_map = {p.property_type: p for p in later.properties}
        common_property_types = set(earlier_property_map.keys()) & set(later_property_map.keys())
        for property_type in common_property_types:
            earlier_property = earlier_property_map[property_type]
            later_property = later_property_map[property_type]
            later_property.conflict = self.diff_combiner.combine_conflicts(
                earlier=earlier_property.conflict, later=later_property.conflict
            )

    def _transfer_relationship_conflicts(
        self,
        earlier: EnrichedDiffRelationship,
        later: EnrichedDiffRelationship,
    ) -> None:
        earlier_elements_by_peer_id = {element.peer_id: element for element in earlier.relationships}
        later_elements_by_peer_id = {element.peer_id: element for element in later.relationships}
        common_peer_ids = set(earlier_elements_by_peer_id.keys()) & set(later_elements_by_peer_id.keys())
        for peer_id in common_peer_ids:
            earlier_element = earlier_elements_by_peer_id[peer_id]
            later_element = later_elements_by_peer_id[peer_id]
            later_element.conflict = self.diff_combiner.combine_conflicts(
                earlier=earlier_element.conflict, later=later_element.conflict
            )
            self._transfer_property_conflicts(earlier=earlier_element, later=later_element)



"""DataLoader for resolving account IDs to full account data."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from aiodataloader import DataLoader

from infrahub.core.constants import SYSTEM_USER_ID, InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.graphql.constants import KIND_GRAPHQL_FIELD_NAME

if TYPE_CHECKING:
    from infrahub.core.branch.models import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase

SYSTEM_ACCOUNT_DISPLAY_LABEL = "[Infrahub System]"
UNKNOWN_ACCOUNT_DISPLAY_LABEL = "[Unknown Account]"


@dataclass
class AccountLoaderParams:
    branch: Branch
    at: Timestamp | None
    fields: dict[str, Any] = field(default_factory=dict)

    def __hash__(self) -> int:
        # Fields are not included in hash as they vary per query but we want to reuse loader
        at_str = self.at.to_string() if self.at else ""
        return hash(f"{at_str}|{self.branch.name}")


class AccountDataLoader(DataLoader[str, dict[str, Any] | None]):
    """DataLoader for batch-loading account data by ID.

    Handles SYSTEM_USER_ID specially by returning synthetic system account data.
    All lookups are branch-agnostic since accounts don't vary by branch.
    """

    def __init__(
        self,
        db: InfrahubDatabase,
        params: AccountLoaderParams,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.db = db
        self.params = params

    def _build_system_account_response(self, fields: dict[str, Any]) -> dict[str, Any]:
        """Build response for SYSTEM_USER_ID based on requested fields."""
        # Always include __kind__ for GraphQL interface type resolution
        response: dict[str, Any] = {KIND_GRAPHQL_FIELD_NAME: InfrahubKind.ACCOUNT}

        for field_name in fields:
            if field_name == "id":
                response["id"] = SYSTEM_USER_ID
            elif field_name == "display_label":
                response["display_label"] = SYSTEM_ACCOUNT_DISPLAY_LABEL
            elif field_name == "__typename":
                response["__typename"] = InfrahubKind.ACCOUNT
            elif field_name == "name":
                response["name"] = {"value": SYSTEM_ACCOUNT_DISPLAY_LABEL}
            elif field_name == "label":
                response["label"] = {"value": SYSTEM_ACCOUNT_DISPLAY_LABEL}
            elif field_name == "description":
                response["description"] = {"value": None}
            else:
                # For any other field, return None
                response[field_name] = None

        return response

    def _build_unknown_account_response(self, account_id: str, fields: dict[str, Any]) -> dict[str, Any]:
        """Build response for an unknown/deleted account based on requested fields."""
        # Always include __kind__ for GraphQL interface type resolution
        response: dict[str, Any] = {KIND_GRAPHQL_FIELD_NAME: InfrahubKind.ACCOUNT}

        for field_name in fields:
            if field_name == "id":
                response["id"] = account_id
            elif field_name == "display_label":
                response["display_label"] = UNKNOWN_ACCOUNT_DISPLAY_LABEL
            elif field_name == "__typename":
                response["__typename"] = InfrahubKind.ACCOUNT
            elif field_name == "name":
                response["name"] = {"value": UNKNOWN_ACCOUNT_DISPLAY_LABEL}
            elif field_name == "label":
                response["label"] = {"value": UNKNOWN_ACCOUNT_DISPLAY_LABEL}
            elif field_name == "description":
                response["description"] = {"value": None}
            else:
                response[field_name] = None

        return response

    async def batch_load_fn(self, keys: list[str]) -> list[dict[str, Any] | None]:
        # Separate system user from real account IDs
        real_account_ids = [k for k in keys if k != SYSTEM_USER_ID]

        # Batch load real accounts
        accounts_by_id: dict[str, dict[str, Any]] = {}
        if real_account_ids:
            async with self.db.start_session(read_only=True) as db:
                nodes = await NodeManager.get_many(
                    db=db,
                    ids=real_account_ids,
                    fields=self.params.fields,
                    at=self.params.at,
                    branch=self.params.branch,
                    branch_agnostic=True,  # Accounts are always branch-agnostic
                )

                for node_id, node in nodes.items():
                    accounts_by_id[node_id] = await node.to_graphql(
                        db=db,
                        fields=self.params.fields,
                    )

        # Build results in same order as keys
        results: list[dict[str, Any] | None] = []
        for key in keys:
            if key == SYSTEM_USER_ID:
                results.append(self._build_system_account_response(self.params.fields))
            elif key in accounts_by_id:
                results.append(accounts_by_id[key])
            else:
                # Account not found - return placeholder
                results.append(self._build_unknown_account_response(key, self.params.fields))

        return results



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class SetCoreProposedChangeCreatedByOnNodeQuery(Query):
    name = "set_coreproposedchange_created_by_on_node"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        MATCH (pc:CoreProposedChange)
        CALL (pc) {
            MATCH (pc)-[r1:IS_RELATED]-(:Relationship {name: "coreaccount__proposedchange_created_by"})-[r2:IS_RELATED]-(account:CoreGenericAccount)
            RETURN account
            ORDER BY r1.from DESC, r2.from DESC
            LIMIT 1
        }
        SET pc.created_by = account.uuid
        """
        self.add_to_query(query)


class Migration060(GraphMigration):
    name: str = "060_set_coreproposedchange_created_by_on_node"
    minimum_version: int = 59
    queries: Sequence[type[Query]] = [SetCoreProposedChangeCreatedByOnNodeQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.migrations.shared import MigrationInput, MigrationResult, get_migration_console
from infrahub.core.query import Query, QueryType
from infrahub.core.schema import GenericSchema, NodeSchema, SchemaRoot, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.log import get_logger

from ..shared import ArbitraryMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()
console = get_migration_console()


@dataclass
class DuplicateSchemaNodeInfo:
    schema_type: str
    name: str
    namespace: str
    keep_uuid: str
    delete_uuid: str

    @property
    def kind(self) -> str:
        return f"{self.namespace}{self.name}"


class FindDuplicateSchemaNodesQuery(Query):
    name: str = "find_duplicate_schema_nodes"
    type: QueryType = QueryType.READ
    insert_return: bool = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict) -> None:  # noqa: ARG002
        query = """
MATCH (default_b:Branch)
WHERE default_b.is_default = TRUE
WITH default_b.name AS default_branch
LIMIT 1

// -------------------
// for every active Schema on the default branch
// -------------------
MATCH (n:SchemaNode|SchemaGeneric)
CALL (n, default_branch) {
    MATCH (n)-[rp:IS_PART_OF]->(root:Root)
    WHERE rp.branch = default_branch
    RETURN rp
    ORDER BY rp.from DESC
    LIMIT 1
}
WITH n, rp, default_branch
WHERE rp.status = "active" AND rp.to IS NULL

// -------------------
// get its name and namespace
// -------------------
CALL (n) {
    MATCH (n)-[r1:HAS_ATTRIBUTE]->(attr:Attribute {name: "name"})-[r2:HAS_VALUE]->(av:AttributeValue)
    WHERE r1.status = "active" AND r2.status = "active"
    RETURN av.value AS name_value
    ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
    LIMIT 1
}
CALL (n) {
    MATCH (n)-[r1:HAS_ATTRIBUTE]->(attr:Attribute {name: "namespace"})-[r2:HAS_VALUE]->(av:AttributeValue)
    WHERE r1.status = "active" AND r2.status = "active"
    RETURN av.value AS namespace_value
    ORDER BY r1.from DESC, r1.status ASC, r2.from DESC, r2.status ASC
    LIMIT 1
}

// -------------------
// get its latest property update time on the default branch
// -------------------
CALL (n, default_branch) {
    MATCH (n)-[:HAS_ATTRIBUTE {branch: default_branch}]->(:Attribute)-[r:HAS_VALUE {branch: default_branch}]->(:AttributeValue)
    WHERE r.status = "active" AND r.to IS NULL
    RETURN max(r.from) AS node_latest
}

// -------------------
// get the latest property update for linked attributes and relationships on the default branch
// -------------------
CALL (n, default_branch) {
    OPTIONAL MATCH (n)-[rc1:IS_RELATED {branch: default_branch}]->(rel:Relationship)
        <-[rc2:IS_RELATED {branch: default_branch}]-(child)
        -[:HAS_ATTRIBUTE {branch: default_branch}]->(:Attribute)
        -[r:HAS_VALUE {branch: default_branch}]->(:AttributeValue)
    WHERE rel.name IN ["schema__node__relationships", "schema__node__attributes"]
    AND rc1.status = "active" AND rc1.to IS NULL
    AND rc2.status = "active" AND rc2.to IS NULL
    AND r.status = "active" AND r.to IS NULL
    RETURN max(r.from) AS child_latest
}

// -------------------
// use the latest Schema, Attribute, or Relationship property update time
// -------------------
WITH n, name_value, namespace_value,
    CASE WHEN child_latest IS NOT NULL AND child_latest > node_latest
        THEN child_latest
        ELSE node_latest
    END AS latest_update,
    CASE WHEN "SchemaNode" IN labels(n) THEN "SchemaNode" ELSE "SchemaGeneric" END AS schema_type
ORDER BY latest_update DESC, elementId(n) DESC
WITH schema_type, name_value, namespace_value, collect(n.uuid) AS uuids
WHERE size(uuids) > 1

// -------------------
// keep the latest update, mark the others to be deleted
// -------------------
UNWIND tail(uuids) AS delete_uuid
RETURN
    schema_type,
    name_value AS name,
    namespace_value AS namespace,
    head(uuids) AS keep_uuid,
    delete_uuid
        """
        self.add_to_query(query)
        self.return_labels = ["schema_type", "name", "namespace", "keep_uuid", "delete_uuid"]

    def get_duplicates(self) -> list[DuplicateSchemaNodeInfo]:
        return [
            DuplicateSchemaNodeInfo(
                schema_type=result.get_as_type("schema_type", return_type=str),
                name=result.get_as_type("name", return_type=str),
                namespace=result.get_as_type("namespace", return_type=str),
                keep_uuid=result.get_as_type("keep_uuid", return_type=str),
                delete_uuid=result.get_as_type("delete_uuid", return_type=str),
            )
            for result in self.get_results()
        ]


class Migration057(ArbitraryMigration):
    name: str = "057_deduplicate_schema_nodes"
    minimum_version: int = 56

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        at = migration_input.at
        user_id = migration_input.user_id

        console.log("[bold]Deduplicating SchemaNode and SchemaGeneric vertices[/bold]")

        # Set up SchemaManager
        manager = SchemaManager()
        registry.schema = manager
        manager.register_schema(schema=SchemaRoot(**internal_schema))
        default_branch = registry.get_branch_from_registry()
        await manager.load_schema_from_db(db=db, branch=default_branch)

        # Find duplicates
        query = await FindDuplicateSchemaNodesQuery.init(db=db)
        await query.execute(db=db)

        duplicates = query.get_duplicates()
        if not duplicates:
            console.log("No duplicate schema nodes found")
            return MigrationResult()

        console.log(f"Found {len(duplicates)} duplicate schema node(s) to remove")

        # Delete older duplicates
        for dup in duplicates:
            console.log(f"Deleting duplicate {dup.schema_type} {dup.kind} (uuid={dup.delete_uuid})")

            obj = await manager.get_one(id=dup.delete_uuid, db=db, branch=default_branch, prefetch_relationships=True)

            schema_obj: NodeSchema | GenericSchema
            if dup.schema_type == "SchemaNode":
                schema_obj = await manager.convert_node_schema_to_schema(schema_node=obj, db=db)
            else:
                schema_obj = await manager.convert_generic_schema_to_schema(schema_node=obj, db=db)

            await manager.delete_node_in_db(db=db, node=schema_obj, user_id=user_id, at=at, branch=default_branch)

        console.log("[bold green]Migration completed successfully[/bold green]")
        return MigrationResult()



from __future__ import annotations

from typing import Any

from infrahub.core.order import OrderModel


def deserialize_order_input(input_data: dict[str, Any] | None) -> OrderModel | None:
    # Corresponds to infrahub.graphql.manager.OrderInput
    if not input_data:
        return None

    order_model = OrderModel(**input_data)
    return order_model



from __future__ import annotations

import ipaddress
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, NamedTuple, get_args

from ..protocols_base import CoreNodeBase
from ..uuidt import UUIDT
from .constants import ATTRIBUTE_METADATA_OBJECT, IP_TYPES, PROPERTIES_FLAG, PROPERTIES_OBJECT, SAFE_VALUE
from .property import NodeProperty

if TYPE_CHECKING:
    from ..schema import AttributeSchemaAPI


class _GraphQLPayloadAttribute(NamedTuple):
    """Result of resolving an attribute value for a GraphQL mutation.

    Attributes:
        payload: Key/value entries to include in the mutation payload
            (e.g. ``{"value": ...}`` or ``{"from_pool": ...}``).
        variables: GraphQL variable bindings for unsafe string values.
        needs_metadata: When ``True``, the payload needs to append property flags/objects
    """

    payload: dict[str, Any]
    variables: dict[str, Any]
    needs_metadata: bool

    def to_dict(self) -> dict[str, Any]:
        return {"data": self.payload, "variables": self.variables}

    def add_properties(self, properties_flag: dict[str, Any], properties_object: dict[str, str | None]) -> None:
        if not self.needs_metadata:
            return
        for prop_name, prop in properties_flag.items():
            self.payload[prop_name] = prop

        for prop_name, prop in properties_object.items():
            self.payload[prop_name] = prop


class Attribute:
    """Represents an attribute of a Node, including its schema, value, and properties."""

    def __init__(self, name: str, schema: AttributeSchemaAPI, data: Any | dict) -> None:
        """
        Args:
            name (str): The name of the attribute.
            schema (AttributeSchema): The schema defining the attribute.
            data (Union[Any, dict]): The data for the attribute, either in raw form or as a dictionary.
        """
        self.name = name
        self._schema = schema
        self._from_pool: dict[str, Any] | None = None

        if isinstance(data, dict) and "from_pool" in data:
            self._from_pool = data.pop("from_pool")
            data.setdefault("value", None)
        elif not isinstance(data, dict) or "value" not in data:
            data = {"value": data}

        self._properties_flag = PROPERTIES_FLAG
        self._properties_object = PROPERTIES_OBJECT
        self._properties = self._properties_flag + self._properties_object

        self._read_only = ["updated_at", "is_inherited"]

        self.id: str | None = data.get("id")

        self._value: Any | None = data.get("value")
        self.value_has_been_mutated = False
        self.is_default: bool | None = data.get("is_default")
        self.is_from_profile: bool | None = data.get("is_from_profile")

        if self._value:
            value_mapper: dict[str, Callable] = {
                "IPHost": ipaddress.ip_interface,
                "IPNetwork": ipaddress.ip_network,
            }
            mapper = value_mapper.get(schema.kind, lambda value: value)
            self._value = mapper(data.get("value"))

        self.is_inherited: bool | None = data.get("is_inherited")
        self.updated_at: str | None = data.get("updated_at")

        self.is_protected: bool | None = data.get("is_protected")

        self.source: NodeProperty | None = None
        self.owner: NodeProperty | None = None
        self.updated_by: NodeProperty | None = None

        for prop_name in self._properties_object:
            if data.get(prop_name):
                setattr(self, prop_name, NodeProperty(data=data.get(prop_name)))  # type: ignore[arg-type]

        for prop_name in ATTRIBUTE_METADATA_OBJECT:
            if data.get(prop_name):
                setattr(self, prop_name, NodeProperty(data=data.get(prop_name)))  # type: ignore[arg-type]

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        self._value = value
        self.value_has_been_mutated = True

    def _initialize_graphql_payload(self) -> _GraphQLPayloadAttribute:
        """Resolve the attribute value into a GraphQL mutation payload object."""

        # Pool-based allocation (dict data or resource-pool node)
        if self._from_pool is not None:
            return _GraphQLPayloadAttribute(payload={"from_pool": self._from_pool}, variables={}, needs_metadata=True)
        if isinstance(self.value, CoreNodeBase) and self.value.is_resource_pool():
            return _GraphQLPayloadAttribute(
                payload={"from_pool": {"id": self.value.id}}, variables={}, needs_metadata=True
            )

        # Null value
        if self.value is None:
            data = {"value": None} if (self._schema.optional and self.value_has_been_mutated) else {}
            return _GraphQLPayloadAttribute(payload=data, variables={}, needs_metadata=False)

        # Unsafe strings need a variable binding to avoid injection
        if isinstance(self.value, str) and not SAFE_VALUE.match(self.value):
            var_name = f"value_{UUIDT.new().hex}"
            return _GraphQLPayloadAttribute(
                payload={"value": f"${var_name}"},
                variables={var_name: self.value},
                needs_metadata=True,
            )

        # Safe strings, IP types, and everything else
        value = self.value.with_prefixlen if isinstance(self.value, get_args(IP_TYPES)) else self.value
        return _GraphQLPayloadAttribute(payload={"value": value}, variables={}, needs_metadata=True)

    def _generate_input_data(self) -> _GraphQLPayloadAttribute:
        """Build the input payload for a GraphQL mutation on this attribute.

        Returns a ResolvedValue object, which contains all the data required.
        """
        graphql_payload = self._initialize_graphql_payload()

        properties_flag: dict[str, Any] = {
            property_name: getattr(self, property_name)
            for property_name in self._properties_flag
            if getattr(self, property_name) is not None
        }
        properties_object: dict[str, str | None] = {
            property_name: getattr(self, property_name)._generate_input_data()
            for property_name in self._properties_object
            if getattr(self, property_name) is not None
        }
        graphql_payload.add_properties(properties_flag, properties_object)

        return graphql_payload

    def _generate_query_data(self, property: bool = False, include_metadata: bool = False) -> dict | None:
        data: dict[str, Any] = {"value": None}

        if property:
            data.update({"is_default": None, "is_from_profile": None})

            for prop_name in self._properties_flag:
                data[prop_name] = None
            for prop_name in self._properties_object:
                data[prop_name] = {"id": None, "display_label": None, "__typename": None}

        if include_metadata:
            data["updated_at"] = None
            for prop_name in ATTRIBUTE_METADATA_OBJECT:
                data[prop_name] = {"id": None, "display_label": None, "__typename": None}

        return data

    def _generate_mutation_query(self) -> dict[str, Any]:
        if self.is_from_pool_attribute():
            # If it points to a pool, ask for the value of the pool allocated resource
            return {self.name: {"value": None}}
        return {}

    def is_from_pool_attribute(self) -> bool:
        """Check whether this attribute's value is sourced from a resource pool.

        Returns:
            True if the attribute value is a resource pool node or was explicitly allocated from a pool.
        """
        return (isinstance(self.value, CoreNodeBase) and self.value.is_resource_pool()) or self._from_pool is not None



from __future__ import annotations

from typing import Any

from pydantic import BaseModel, model_validator

CONVERT_OBJECT_MUTATION = """
    mutation($node_id: String!, $target_kind: String!, $fields_mapping: GenericScalar!) {
        ConvertObjectType(data: {
                node_id: $node_id,
                target_kind: $target_kind,
                fields_mapping: $fields_mapping
            }) {
                ok
                node
        }
    }
"""


class ConversionFieldValue(BaseModel):  # Only one of these fields can be not None
    """
    Holds the new value of the destination field during an object conversion.
    Use `attribute_value` to specify the new raw value of an attribute.
    Use `peer_id` to specify new peer of a cardinality one relationship.
    Use `peers_ids` to specify new peers of a cardinality many relationship.
    Only one of `attribute_value`, `peer_id` and `peers_ids` can be specified.
    """

    attribute_value: Any | None = None
    peer_id: str | None = None
    peers_ids: list[str] | None = None

    @model_validator(mode="after")
    def check_only_one_field(self) -> ConversionFieldValue:
        fields = [self.attribute_value, self.peer_id, self.peers_ids]
        set_fields = [f for f in fields if f is not None]
        if len(set_fields) != 1:
            raise ValueError("Exactly one of `attribute_value`, `peer_id`, or `peers_ids` must be set")
        return self


class ConversionFieldInput(BaseModel):
    """
    Indicates how to fill in the value of the destination field during an object conversion.
    Use `source_field` to reuse the value of the corresponding field of the object being converted.
    Use `data` to specify the new value for the field.
    Use `use_default_value` to set the destination field to its schema default.
    Only one of `source_field`, `data`, or `use_default_value` can be specified.
    """

    source_field: str | None = None
    data: ConversionFieldValue | None = None
    use_default_value: bool = False

    @model_validator(mode="after")
    def check_only_one_field(self) -> ConversionFieldInput:
        fields_set = [self.source_field is not None, self.data is not None, self.use_default_value is True]
        if sum(fields_set) != 1:
            raise ValueError("Exactly one of `source_field`, `data` or `use_default_value` must be set")
        return self



from typing import Any

from fastapi.responses import JSONResponse
from pydantic import ValidationError

from infrahub.exceptions import Error


async def generic_api_exception_handler(_: Any, exc: Exception, http_code: int = 500) -> JSONResponse:
    """Generic API Exception handler."""
    if isinstance(exc, Error):
        if exc.HTTP_CODE:
            http_code = exc.HTTP_CODE
        if isinstance(exc.errors, list):
            messages = exc.errors
        else:
            messages = [str(exc.message) if exc.message else exc.DESCRIPTION]
    elif isinstance(exc, ValidationError):
        messages = [ed["msg"] for ed in exc.errors()]
    else:
        messages = [str(exc)]
    error_dict = {
        "data": None,
        "errors": [{"message": message, "extensions": {"code": http_code}} for message in messages],
    }

    return JSONResponse(status_code=http_code, content=error_dict)
